#include "struct.h"
#include "sysvars.h"
#include "ops.h"

#include <cstring>

bool Bot::docode() {
	// handle venom
	for (std::list<VenomData>::iterator p = venomed.begin(); p != venomed.end(); ++p) {
		data[p->loc-1] = p->val;
		if (p->poison) data[s_paralyzed] = 1;
		else data[s_poisoned] = 1;
	}

	code->code_exec(*this);
	const float aim = db2rad((12560-data[s_setaim]+data[s_aimdx]-data[s_aimsx])%1256);

	// rotate
	b_body->SetXForm(b_body->GetPosition(), aim);

	// fix if requested
	if (fixed != (data[s_fixpos] > 0)) {
		fixed = data[s_fixpos] > 0;
		ResetMass();
	}

	create_substances();

	// determine movement
	if (!fixed) {
		move();
	}

	// shoot
	shoot();

	// feed veggies
	if (veg) switch (sim->prefs.vegfeed_type) {
		default:
		case NRG_PER_CYCLE:
			nrg += sim->prefs.vegfeed;
			break;
		case NRG_PER_KILOBODY:
			nrg += sim->prefs.vegfeed * body * 0.001f;
			break;
	}

	// apply generic costs
	nrg -= body*sim->prefs.bodyupkeepcost;
	nrg -= sim->prefs.agecost;

	// cap energy
	if (nrg > 32000.0f) nrg = 32000.0f;

	// reproduce
	repro();

	if (radius_changed) {
		radius = getradius(body);
		b_body->DestroyShape(b_shape);
		b2CircleDef sdef;
		sdef.friction = sim->prefs.friction;
		sdef.restitution = sim->prefs.elast;
		sdef.radius = radius*0.01f;
		sdef.localPosition.SetZero();
		sdef.filter.groupIndex = 4;
		b_shape = b_body->CreateShape(&sdef);
		radius_changed = false;
	}

/*		if (pos.x*100-temp.radius <= 0) pos.x = temp.radius*0.01;
	if (pos.x*100+temp.radius >= width) pos.x = (width-temp.radius)*0.01;
	if (pos.y*100-temp.radius <= 0) pos.y = temp.radius*0.01;
	if (pos.y*100+temp.radius >= height) pos.y = (height-temp.radius)*0.01;*/

	// kill dead bots
	if (nrg <= 0.0f || body <= 0.0f) {
		return false;
	}

	return true;
}

void Bot::create_substances() {
	// create substances
	// disallow creating too much substance
	if (poison+data[s_strpoison] > 32000) data[s_strpoison] = 32000-(int) poison;
	if (venom+data[s_strvenom] > 32000) data[s_strvenom] = 32000-(int) venom;
	if (shell+data[s_mkshell] > 32000) data[s_mkshell] = 32000-(int) shell;
	if (slime+data[s_mkslime] > 32000) data[s_mkslime] = 32000-(int) slime;

	poison += data[s_strpoison];
	nrg -= data[s_strpoison]*sim->prefs.poisoncost;

	venom += data[s_strvenom];
	nrg -= data[s_strvenom]*sim->prefs.venomcost;

	shell += data[s_mkshell];
	nrg -= data[s_mkshell]*sim->prefs.shellcost;

	slime += data[s_mkslime];
	nrg -= data[s_mkslime]*sim->prefs.slimecost;

	slime *= 0.98f; // degrade slime
	if (slime < 0.5f) slime = 0; // to prevent rounding badness

	// convert body - rate: 10 nrg/body
	float mkbody = data[s_strbody]-data[s_fdbody];
	// cap creation at 100 body (1000 nrg)
	if (mkbody < -1000.0f) mkbody = -1000.0f;
	if (mkbody > 1000.0f) mkbody = 1000.0f;
	// don't let body go over 32000
	if (body+mkbody*0.1f > 32000.0f) mkbody = (32000.0f-body)*10.0f;
	if (nrg-mkbody > 32000.0f) mkbody = 32000.0f-nrg;
	if (mkbody) {
		body += mkbody*0.1f;
		nrg -= mkbody;
		radius_changed = true;
	}

	if (data[s_mkshell] || mkbody)
		ResetMass();
}

void Bot::move() {
	const float aim = b_body->GetAngle(),
		up = (data[s_up]-data[s_dn])*0.005f,
		left = (data[s_sx]-data[s_dx])*0.005f;
	const b2Vec2 pos = b_body->GetPosition();

	if (fabs(up) > 0.0f || fabs(left) > 0.0f) {
		// apply movement cost
		nrg -= (fabs(up)+fabs(left))*sim->prefs.movcost;

		// apply forces
		b2Vec2 vect(cos(aim)*up+sin(aim)*left, sin(aim)*up-cos(aim)*left);
		vect *= sim->prefs.move_eff;
		b_body->ApplyForce(vect, pos);
	}

	// Brownian motion, ish - helps prevent bots from getting crushed at walls
//	if (rand() & 1) {
//		float32 invmass = 1.0f/(temp->Mass()+10);
//		pos.x += (static_cast<float32>(rand())*0.08f/RAND_MAX-0.04f)*invmass;
//		pos.y += (static_cast<float32>(rand())*0.08f/RAND_MAX-0.04f)*invmass;
//	}
}

void Bot::repro() {
	if (data[s_repro] > 0) {
		const float aim = b_body->GetAngle();
		// limit number of bots
		if ((veg && sim->numveg >= sim->prefs.maxveg) || (!veg && sim->bots.size()-sim->numveg >= sim->prefs.maxnonveg)) {
			return;
		}
		// data[s_repro] is a percentage, convert to float
		float repro = static_cast<float>(data[s_repro])*0.01f;
		if (repro > 1.0f) repro = 1.0f;

		if (body*repro < 2.0f || nrg*repro < 2.0f) {
			// offspring not healthy enough
			return;
		}
		float new_radius = getradius(body*repro)*0.01f;
		//if (new_radius <= 0) {
		//	// shouldn't happen
		//	std::cerr << "new radius would be less than zero? " << new_radius << std::endl;
		//	goto repro_fail;
		//}

		// determine new offspring location
		b2Vec2 tpos = b_body->GetPosition();
		tpos.x += cos(aim)*(radius*0.01f+new_radius+0.05f);
		tpos.y += sin(aim)*(radius*0.01f+new_radius+0.05f);
		// ensure that there is enough space for the offspring
		if (tpos.x-new_radius < 0.0f || tpos.x+new_radius > sim->prefs.width*0.01f ||
				tpos.y-new_radius < 0.0f || tpos.y+new_radius > sim->prefs.height*0.01f) {
			// out of bounds
			return;
		}
		// determine possible intersectors with the offspring
		b2AABB aabb;
		aabb.lowerBound.Set(tpos.x-new_radius, tpos.y-new_radius);
		aabb.upperBound.Set(tpos.x+new_radius, tpos.y+new_radius);
		b2Shape* buffer[sim->bots.size()+sim->numshots];
		int count = sim->b_world->Query(aabb, buffer, sim->bots.size()+sim->numshots);
		for (int j = count; j--; ) {
			//SimObject* temp2 = (SimObject*) buffer[j]->GetBody()->GetUserData();
			//if (temp2 == NULL || temp2->type != BOT) continue;
			if (buffer[j]->GetFilterData().groupIndex != 4) continue;
			Bot& temp2 = *((Bot*) buffer[j]->GetBody()->GetUserData());
			b2Vec2 dist = tpos-(buffer[j]->GetBody()->GetPosition());
			float tradius = temp2.radius*0.01f+new_radius;
			if (dist.LengthSquared() < tradius*tradius) {
				// bots would intersect
				return;
			}
		}
		// create offspring
		size_t id = sim->add_bot(tpos.x*100.0f, tpos.y*100.0f, fmod(aim+M_PI, M_PI*2.0f), body*repro, code);
		Bot& offspring = sim->bots.at(id);
		offspring.nrg = nrg*repro;
		offspring.color[0] = color[0];
		offspring.color[1] = color[1];
		offspring.color[2] = color[2];
		offspring.name = name;
		offspring.veg = veg;
		offspring.fixed = fixed;
		if (offspring.fixed) std::cerr << "offspring fixed\n";
		offspring.novis = novis;
		offspring.noexec = noexec;
		offspring.ResetMass();
		if (veg) ++sim->numveg;
		// reduce parent resources
		nrg *= 1.0f-repro;
		body *= 1.0f-repro;
		radius_changed = true;
	}
}

void Bot::vision() {
	// reset eye data
	memset(eyes, 0, sizeof(eyes));
	if (!novis) {
		// determine bots in the vicinity
		const b2Vec2 pos = b_body->GetPosition();
//		b2Shape* buffer[sim->numbots+sim->numshots+4];
//		b2AABB aabb;
		float maxdist;
		for (int j = 0; j < 9; ++j) {
			const float width = 0.17453292519943295769f+db2rad(data[s_eye1width+j]);
			const float thisdist = width_to_range((width<0.1f)?0.1f:width);
			if (thisdist > maxdist) maxdist = thisdist;
		}
		if (maxdist <= 1.0f) maxdist = 1.0f;
		maxdist += radius;
		maxdist *= 0.01f;
		for (size_t i = sim->bots.size(); i--; ) {
			if (i == id) continue;
			Bot& temp = sim->bots.at(i);
			const float tmpdist = maxdist + temp.radius*0.01f;
			if ((temp.b_body->GetPosition()-pos).LengthSquared() < tmpdist*tmpdist) {
				see(temp);
			}
		}
	}

	data[s_eye1] = eyes[0].eyeval;
	data[s_eye2] = eyes[1].eyeval;
	data[s_eye3] = eyes[2].eyeval;
	data[s_eye4] = eyes[3].eyeval;
	data[s_eye5] = eyes[4].eyeval;
	data[s_eye6] = eyes[5].eyeval;
	data[s_eye7] = eyes[6].eyeval;
	data[s_eye8] = eyes[7].eyeval;
	data[s_eye9] = eyes[8].eyeval;
	fill_refvars();
}

void Bot::fill_refvars() {
	int focuseye = (data[s_focuseye]+4) % 9;
	if (focuseye < 0) focuseye += 9;
	if (eyes[focuseye].eyeval) {
		data[s_refxpos] = eyes[focuseye].refxpos;
		data[s_refypos] = eyes[focuseye].refypos;
		data[s_refnrg] = eyes[focuseye].refnrg;
		data[s_refbody] = eyes[focuseye].refbody;
		data[s_refshell] = eyes[focuseye].refshell;
		data[s_refpoison] = eyes[focuseye].refpoison;
		data[s_refvenom] = eyes[focuseye].refvenom;
		data[s_refveldn] = eyes[focuseye].refveldn;
		data[s_refvelup] = data[s_refvel] = eyes[focuseye].refvelup;
		data[s_refvelsx] = eyes[focuseye].refvelsx;
		data[s_refveldx] = eyes[focuseye].refveldx;
		data[s_refvelscalar] = eyes[focuseye].refvelscalar;
		data[s_refeye] = eyes[focuseye].refeye;
		data[s_refshoot] = eyes[focuseye].refshoot;
		data[s_in1] = eyes[focuseye].in1;
		data[s_in2] = eyes[focuseye].in2;
		data[s_in3] = eyes[focuseye].in3;
		data[s_in4] = eyes[focuseye].in4;
		data[s_in5] = eyes[focuseye].in5;
		data[s_in6] = eyes[focuseye].in6;
		data[s_in7] = eyes[focuseye].in7;
		data[s_in8] = eyes[focuseye].in8;
		data[s_in9] = eyes[focuseye].in9;
		data[s_in10] = eyes[focuseye].in10;
		data[s_memval] = eyes[focuseye].memval;
	} else {
		data[s_refxpos] = 0;
		data[s_refypos] = 0;
		data[s_refnrg] = 0;
		data[s_refbody] = 0;
		data[s_refshell] = 0;
		data[s_refpoison] = 0;
		data[s_refvenom] = 0;
		data[s_refvel] = 0;
		data[s_refvelup] = 0;
		data[s_refveldn] = 0;
		data[s_refveldx] = 0;
		data[s_refvelsx] = 0;
		data[s_refeye] = 0;
		data[s_refshoot] = 0;
		data[s_in1] = 0;
		data[s_in2] = 0;
		data[s_in3] = 0;
		data[s_in4] = 0;
		data[s_in5] = 0;
		data[s_in6] = 0;
		data[s_in7] = 0;
		data[s_in8] = 0;
		data[s_in9] = 0;
		data[s_in10] = 0;
		data[s_memval] = 0;
	}
}

// represents an angle from a to b
class AngRange {
	float a, b;
public:
	AngRange(float _a, float _b) : a(fmod(_a+M_PI*2.0f, M_PI*2.0f)), b(fmod(_b+M_PI*2.0f, M_PI*2.0f)) {
		if (b < a) b += M_PI*2.0f;
	}
	bool in(float ang) const {
		// normalize values
//		ang = fmod(ang+M_PI*2.0f, M_PI*2.0f);
//		return (ang > a && ang < b) || (ang+M_PI*2.0f > a && ang+M_PI*2.0f < b);
		return ang > a && ang < b;
	}
	bool in(const AngRange& range) const {
		return in(range.a) || in(range.b) || range.in(a) || range.in(b);
	}
};

// shamelessly stolen from http://www.phatcode.net/articles.php?id=459
/*bool circle_in_triangle(float cx, float cy, float r, float v1x, float v1y, float v2x, float v2y, float v3x, float v3y) {
	// TEST 1: Vertex within circle
	const float c1x = cx-v1x;
	const float c1y = cy-v1y;

	const float radiusSqr = r*r;
	const float c1sqr = c1x*c1x+c1y*c1y-radiusSqr;

	if (c1sqr <= 0.0f) return true;

	const float c2x = cx-v2x;
	const float c2y = cy-v2y;
	const float c2sqr = c2x*c2x+c2y*c2y-radiusSqr;

	if (c2sqr <= 0.0f) return true;

	const float c3x = cx-v3x;
	const float c3y = cy-v3y;
	const float c3sqr = c3x*c3x+c3y*c3y-radiusSqr;

	if (c3sqr <= 0.0f) return true;

	// TEST 2: Circle centre within triangle

	// Calculate edges
	const float e1x = v2x-v1x;
	const float e1y = v2y-v1y;

	const float e2x = v3x-v2x;
	const float e2y = v3y-v2y;

	const float e3x = v1x-v3x;
	const float e3y = v1y-v3y;

	if (e1y*c1x >= e1x*c1y && e2y*c2x >= e2x*c2y && e3y*c3x >= e3x*c3y) return true;

	// TEST 3: Circle intersects edge
	float k = c1x*e1x+c1y*e1y;

	if (k > 0.0f) {
		const float len = e1x*e1x + e1y*e1y;     // squared len
		if (k < len && c1sqr*len <= k*k) return true;
	}

	// Second edge
	k = c2x*e2x+c2y*e2y;

	if (k > 0.0f) {
		const float len = e2x*e2x+e2y*e2y;
		if (k < len && c2sqr*len <= k*k) return true;
	}

	// Third edge
	k = c3x*e3x+c3y*e3y;

	if (k > 0.0f) {
		const float len = e3x*e3x+e3y*e3y;
		if (k < len && c3sqr*len <= k*k) return true;
	}

	return false;
}

bool circle_in_triangle(long cx, long cy, long r, long v1x, long v1y, long v2x, long v2y, long v3x, long v3y) {
	// TEST 1: Vertex within circle
	const long c1x = cx-v1x;
	const long c1y = cy-v1y;
	const long radiusSqr = r*r;
	const long c1sqr = c1x*c1x+c1y*c1y-radiusSqr;
	if (c1sqr <= 0) return true;

	const long c2x = cx-v2x;
	const long c2y = cy-v2y;
	const long c2sqr = c2x*c2x+c2y*c2y-radiusSqr;
	if (c2sqr <= 0) return true;

	const long c3x = cx-v3x;
	const long c3y = cy-v3y;
	const long c3sqr = c3x*c3x+c3y*c3y-radiusSqr;
	if (c3sqr <= 0) return true;

	// TEST 2: Circle centre within triangle

	// Calculate edges
	const long e1x = v2x-v1x;
	const long e1y = v2y-v1y;
	const long e2x = v3x-v2x;
	const long e2y = v3y-v2y;
	const long e3x = v1x-v3x;
	const long e3y = v1y-v3y;
	if (e1y*c1x >= e1x*c1y && e2y*c2x >= e2x*c2y && e3y*c3x >= e3x*c3y) return true;

	// TEST 3: Circle intersects edge
	long k = c1x*e1x+c1y*e1y;
	if (k > 0) {
		const long len = e1x*e1x + e1y*e1y;     // squared len
		if (k < len && c1sqr*len <= k*k) return true;
	}

	// Second edge
	k = c2x*e2x+c2y*e2y;
	if (k > 0) {
		const long len = e2x*e2x+e2y*e2y;
		if (k < len && c2sqr*len <= k*k) return true;
	}

	// Third edge
	k = c3x*e3x+c3y*e3y;
	if (k > 0) {
		const long len = e3x*e3x+e3y*e3y;
		if (k < len && c3sqr*len <= k*k) return true;
	}

	return false;
}*/

// Determines eyesight
void Bot::see(const Bot& b) {
	// insert huge bottleneck here
	b2Vec2 pos = 100.0f*b_body->GetPosition(), bpos = 100.0f*b.b_body->GetPosition();
	b2Vec2 vdist = bpos - pos;

	const float dist = vdist.Length();
	//if (dist > 1440.0f*1440.0f) return; // sight range 1440
	const float edist = dist-b.radius-radius,
		/*tmp = (edist+10.0f)/1440.0f,*/
		ang = fmod(atan2f(vdist.y, vdist.x)+M_PI*2.0f, M_PI*2.0f), // angle
		halfangrad = asinf(b.radius/dist); // the amount of "eye space" percieved
	const AngRange angrange(ang-halfangrad, ang+halfangrad);

	const int eyeval = (edist <= 0.1f) ? 32000 : static_cast<int>(2073600.0f/((edist+10.0f)*(edist+10.0f))); // eye value
					 // 32000 when bots are touching or nearly touching

	EyeInfo eyedata;
	eyedata.id = b.id;
	eyedata.eyeval = eyeval;
	float x = bpos.x, y = bpos.y;
	field2bot(x, y);
	eyedata.refxpos = static_cast<int>(x);
	eyedata.refypos = static_cast<int>(y);
	eyedata.refnrg = static_cast<int>(b.nrg+0.5f);
	eyedata.refbody = static_cast<int>(b.body+0.5f);
	eyedata.refshell = static_cast<int>(b.shell+0.5f);
	eyedata.refpoison = static_cast<int>(b.poison+0.5f);
	eyedata.refvenom = static_cast<int>(b.venom+0.5f);
	{
		b2Vec2 rvel = 100.0f*(b.b_body->GetLinearVelocity()-b_body->GetLinearVelocity());
		const float ang = atan2(rvel.y, rvel.x)-b_body->GetAngle();
		const float mag = rvel.Length();
		rvel.x = cos(ang)*mag;
		rvel.y = sin(ang)*mag;
		eyedata.refveldn = -rvel.x;
		eyedata.refvelup = rvel.x;
		eyedata.refvelsx = -rvel.y;
		eyedata.refveldx = rvel.y;
		eyedata.refvelscalar = mag;
		//if (sim->bm && id == sim->bmid) std::cout << "velup " << rvel.x << " veldx " << rvel.y << std::endl;
	}
	eyedata.refeye = b.code->myeye;
	eyedata.refshoot = b.code->myshoot;
	eyedata.in1 = b.data[s_out1];
	eyedata.in2 = b.data[s_out2];
	eyedata.in3 = b.data[s_out3];
	eyedata.in4 = b.data[s_out4];
	eyedata.in5 = b.data[s_out5];
	eyedata.in6 = b.data[s_out6];
	eyedata.in7 = b.data[s_out7];
	eyedata.in8 = b.data[s_out8];
	eyedata.in9 = b.data[s_out9];
	eyedata.in10 = b.data[s_out10];
	eyedata.memval = b.data[data[s_memloc]-1];

	// a seeing b
	for (int i = 0; i < 9; ++i) {
		if (eyeval < eyes[i].eyeval) continue;
		const float eyeang = b_body->GetAngle()+((float) i)*0.17453292519943295769f-0.69813170079773183076f-db2rad(data[s_eye1dir+i]);
		float eyewidth = 0.17453292519943295769f+db2rad(data[s_eye1width+i]);
		if (eyewidth < 0.1f) eyewidth = 0.1f;
		const float halfeyewidth = eyewidth*0.5f;
		const float range = width_to_range(eyewidth);
		if (edist > range) continue;
		if (angrange.in(AngRange(eyeang-halfeyewidth, eyeang+halfeyewidth))) {//fill_eye(b.id, i, eyevaln);
//		const float cos1 = cos(eyeang-halfeyewidth), sin1 = sin(eyeang-halfeyewidth),
//				cos2 = cos(eyeang+halfeyewidth), sin2 = sin(eyeang+halfeyewidth);
//		if (circle_in_triangle((long) bpos.x, (long) bpos.y, (long) b.radius, (long) pos.x, (long) pos.y, (long) (cos1*range+pos.x),
//				(long) (sin1*range+pos.y), (long) (cos2*range+pos.x), (long) (sin2*range+pos.y))) {
//		if (circle_in_triangle(bpos.x, bpos.y, b.radius, pos.x, pos.y, (cos1*range+pos.x), (sin1*range+pos.y), (cos2*range+pos.x), (sin2*range+pos.y))) {
			eyes[i] = eyedata;
		}
	}

	// b seeing a
/*	{
		const float rang = fmod(ang-b.b_body->GetAngle()+M_PI*4, M_PI*2)-M_PI, // relative angle
			angrad = radius/(dist-b.radius),
			rangmin = rang-angrad,
			rangmax = rang+angrad;
		if (rangmin <= M_PI*0.25 || rangmax >= M_PI*1.75) {
			// 90 degree vision
			int eyemin = static_cast<int>(rangmin*9/M_PI+4.5);
			int eyemax = static_cast<int>(rangmax*9/M_PI+4.5);
			// 0-1 = eye1, ... 8-9 = eye9
			if (eyemin < 0) eyemin = 0;
			if (eyemax > 8) eyemax = 8;
			for (int eyen = eyemin; eyen <= eyemax; ++eyen) {
				if (b.eyeval[eyen] < eyevaln) {
					b.fill_eye(id, eyen, eyevaln);
				}
			}
		}
	}*/
}

void Bot::apply_venom(int loc, int val, int shootval, bool poison) {
	// record venom
	bool added = false;
	unsigned int time = (unsigned int) pow((float) shootval, 0.75);
	for (std::list<VenomData>::iterator p = venomed.begin(); p != venomed.end(); ++p) {
		if (p->loc == loc && p->poison == poison) {
			if (p->val == val) {
				// add to the effect of the venom
				p->time += time;
			} else {
				// venom replaces old venom
				p->val = val;
				p->time = time;
			}
			added = true;
			break;
		}
	}
	if (!added) {
		// new venom
		VenomData data;
		data.loc = loc;
		data.val = val;
		data.time = time;
		data.poison = poison;
		venomed.push_back(data);
	}
}

bool Bot::hit(Shot& shot) {
	if (shot.age >= 10000000 || shot.parent == this) return false; // shot already processed or came from this bot
	if (shot.shoot > 0) {
		data[shot.shoot-1] = shot.val;
		shot.nrg = 0;
		shot.age = 10000000;
	} else if (shot.shoot < 0) {
		// replaces the per-cycle arctan
		shot.nrg *= atan((1.0f-((float) shot.age)/((float) shot.range))*15.0f)*0.6647927652098616;
		switch (shot.shoot) {
			case -1: {
				// nrg feeding shot
				const float amt = shot.val*shot.nrg/(shot.range*40); // FIXME: should take "EnergyProp" option FIXME: simplify this mess
				if (poison > amt) {
					// release poison
					poison -= amt*0.9f;
					shot.shoot = -5;
					shot.val = amt;
					shot.b_body->SetLinearVelocity(-1.0f*shot.b_body->GetLinearVelocity()); // reverse direction
					shot.color[0] = shot.color[1] = 1.0f;
					shot.color[2] = 0.0f;
					shot.age = 0;
				} else {
					nrg -= amt;
					// convert shot to nrg
					shot.shoot = -2;
					shot.nrg = amt*0.9f; // 90% efficiency, FIXME: should be an option
					shot.b_body->SetLinearVelocity(-0.5f*shot.b_body->GetLinearVelocity()); // reverse direction
					shot.color[0] = shot.color[1] = shot.color[2] = 1.0f; // color = white
					shot.age = 0; // reset age
					shot.range *= 2;
					if (shot.parent != NULL && nrg < 0.0f) ++shot.parent->kills;
				}
				shot.reparent(*this);
				break;
			}
			case -2: {
				// nrg giving shot
				if (shot.nrg+nrg > 32000.0f) {
					body += (shot.nrg+nrg-32000.0f)*0.1f;
					if (body > 32000.0f) body = 32000.0f;
					shot.nrg = 32000.0f-nrg;
					radius_changed = true;
				}
				nrg += shot.nrg;
				if (nrg > 32000.0f) nrg = 32000.0f;
				shot.nrg = 0;
				shot.age = 10000000;
				break;
			}
			case -3: {
				// venom shot
				apply_venom(shot.loc, shot.vval, shot.val, false);
				shot.nrg = 0;
				shot.age = 10000000;
				break;
			}
			case -4: {
				// waste shot
				waste += shot.val;
				if (waste > 32000.0f) waste = 32000.0f;
				shot.nrg = 0;
				shot.age = 10000000;
				break;
			}
			case -5: {
				// poison shot
				apply_venom(shot.loc, 0, shot.val, true);
				shot.nrg = 0;
				shot.age = 10000000;
				break;
			}
			case -6: {
				// body-feeding shot
				float amt = shot.val*shot.nrg/(shot.range*40);
				body -= amt;
				radius_changed = true;
				if (shot.parent != NULL && body < 0.0f) ++shot.parent->kills;
				// convert shot to nrg
				shot.shoot = -2;
				shot.nrg = amt*9.0f; // 90% efficiency, FIXME: should be an option
				shot.b_body->SetLinearVelocity(-0.5f*shot.b_body->GetLinearVelocity()); // reverse direction
				shot.color[0] = shot.color[1] = shot.color[2] = 1.0f; // color = white
				shot.age = 0;
				shot.range *= 2;
				shot.reparent(*this);
				break;
			}
			case -8: {
				// reproduction shot
				// FIXME: implement this
				break;
			}
		}
	}
	return true;
}

// bot shooting
void Bot::shoot() {
	data[s_shoot] = -((-data[s_shoot] - 1) % 8 + 1); // -9 -> -1, -10 -> -2, etc.
	if (data[s_shoot] != 0 && data[s_shoot] != -5) { // poison disallowed
		float aim = b_body->GetAngle();
		if (data[s_aimshoot]) {
			aim += db2rad(1256-data[s_aimshoot]);
		} else if (data[s_backshot]) {
			aim += M_PI;
		}
		aim += randf(-0.05, 0.05);
		// determine shot position
		b2Vec2 spos(cos(aim), sin(aim));
		spos *= radius*0.01f+0.1f;
		spos += b_body->GetPosition();

		// shot within boundaries?
		if (spos.x < static_cast<float>(sim->prefs.width)*0.01f && spos.x > 0 &&
				spos.y < static_cast<float>(sim->prefs.height)*0.01f && spos.y > 0) {
			// shot velocity
			b2Vec2 vel(cos(aim)*1.6f, sin(aim)*1.6f);

			// add in bot velocity (they are being fired)
			vel += b_body->GetLinearVelocity();
			float costmult = 1.0f;
			/*size_t shotid = sim->add_shot(spos.x, spos.y); // create shot
			Shot& stemp = sim->shots.at(shotid);
			if (!shots.empty()) {
				shots.insert(--shots.end(), shotid);
			} else {
				shots.insert(shotid);
			}*/
			Shot& stemp = sim->add_shot(*this, spos.x, spos.y);

			stemp.range = 1.0f;
			switch (data[s_shoot]) {
				case -1:
				case -6:
					stemp.val = (int) (body*0.2f)+20;
					// body/nrg-feed shot
					if (data[s_shootval] < 0) {
						// increase range
						stemp.range = log2(2-data[s_shootval]);
						costmult = 1-data[s_shootval];
					} else if (data[s_shootval] > 0) {
						// increase power
						stemp.val *= log2(data[s_shootval]+2);
						costmult = data[s_shootval]+1;
					}
					break;
				case -2:
					// energy shot
					if (data[s_shootval] < 0) data[s_shootval] = 1; // cheater
					stemp.val = (float) data[s_shootval];
					if (stemp.val > nrg) stemp.val = nrg;
					nrg -= stemp.val;
					break;
				case -3:
					// venom shot
					if (data[s_shootval] < 0) data[s_shootval] = 1; // cheater
					stemp.val = (float) data[s_shootval];
					if (stemp.val > venom) stemp.val = venom;
					venom -= stemp.val;
					stemp.loc = data[s_vloc];
					stemp.val = data[s_venval];
					break;
				case -4:
					// waste shot
					if (data[s_shootval] < 0) data[s_shootval] = 1; // cheater
					stemp.val = (float) data[s_shootval];
					if (stemp.val > waste) stemp.val = waste;
					waste -= data[s_shootval];
					break;
				default:
					if (data[s_shoot] > 1000) {
						// 1001 -> 1, 1002 -> 2, etc.
						--data[s_shoot];
						data[s_shoot] %= 1000;
						++data[s_shoot];
					}
					stemp.val = (float) data[s_shootval];
			}
			if (data[s_shoot] == -2) {
				// white nrg shots
				stemp.color[0] = stemp.color[1] = stemp.color[2] = 1.0f;
				// nrg = value
				stemp.nrg = stemp.val;
			} else {
				// new shot has same colour as shooting bot
				stemp.color[0] = color[0];
				stemp.color[1] = color[1];
				stemp.color[2] = color[2];
				// nrg stuff
				if (body > 10) {
					int temp = ((int) (log(body)*60.0f*stemp.range/40.0f))+1;
					stemp.range = temp;
					stemp.nrg = temp*40;
				} else {
					stemp.nrg = stemp.range*40;
				}
			}
			// set shot data
			stemp.shoot = data[s_shoot];
			stemp.body = body;
/*			stemp.parent = id;
			stemp.parent_exists = true;*/
			// set velocity
			stemp.b_body->SetLinearVelocity(vel);
			// apply cost
			nrg -= sim->prefs.shotcost*costmult;
		}
	}
	data[s_shoot] = 0;
	data[s_shootval] = 0;
	data[s_aimshoot] = 0;
	data[s_backshot] = 0;
}

// adds a bot to the sim
Bot::Bot(Sim* _sim, float x, float y, float ang, float _body, Code::ptr _code) : code(_code), sim(_sim),
		body(_body), radius(getradius(_body)) {
	data[s_aim] = data[s_setaim] = 1256-rad2db(ang);
	// zero fields
	shell = 0;
	slime = 0;
	poison = 0;
	venom = 0;
	vtimer = 0;
	waste = 0;
	robage = 0;
	fixed = false;
	veg = false;
	memset(data, 0, sizeof(data));
	radius_changed = false;
	// create body
	b2BodyDef bdef;
	bdef.position.Set(x*0.01f, y*0.01f);
	bdef.angle = ang;
	bdef.allowSleep = true;
	bdef.isSleeping = false;
	bdef.isBullet = true;
	bdef.fixedRotation = true;
	bdef.userData = this;
	bdef.linearDamping = sim->prefs.friction;
	bdef.angularDamping = 100.0f;
	b_body = sim->b_world->CreateBody(&bdef);
	// create shape
	b2CircleDef sdef;
	sdef.friction = sim->prefs.friction;
	sdef.restitution = sim->prefs.elast;
	sdef.radius = radius*0.01f;
	sdef.localPosition.Set(0.0f, 0.0f);
	sdef.filter.groupIndex = 4; // bots have +4, shots -4
	b_shape = b_body->CreateShape(&sdef);
	// etc.
	ResetMass();
}

Bot::~Bot() {
	/*for (std::set<size_t>::iterator q = shots.begin(); q != shots.end(); ++q) {
		try {
			Shot& shot = sim->shots.at(*q);
			if (shot.parent == id) {
				shot.parent_exists = false;
			} else {
				std::cerr << "!! shot #" << *q << " has different parent !!";
			}
		} catch (...) {
			std::cerr << "!! shot #" << *q << " not in shot array !!";
		}
	}*/
	for (shotlist::iterator p; (p = shots.begin()) != shots.end();) {
		Shot& shot = *p;
		// orphaned shot! how sad.
		assert(shot.is_linked());
		assert(shot.parent == this);
		shot.unlink();
		shot.parent = NULL;
		sim->orphan_shots.push_back(shot);
	}
	sim->b_world->DestroyBody(b_body);
	/*if (sim->mjoint != NULL && mjoint->GetBody2()->GetUserData() == this) mjoint = NULL;
	if (veg) --numveg;
	--numbots;
	if (bm && bmid == bot.id) {
		bm = false;
		down = false;
	}*/
}
