#include "struct.h"

#include <map>
#include <string>
#include <fstream>
#include <sstream>
#include <cstring>
#include <cstdlib>

#include "sysvars.h"
//#include "gperf_sysvars.h"
#include "ops.h"

// finds op bytecode
int getop(const std::string& name) {
	for (unsigned int i = num_ops; i--;) {
		if (name == ops[i].name) {
			return ops[i].num;
		}
	}
	return -1;
}

class sysvarkeeper {
	std::map<std::string, int> customdefs;
public:
	void add_custom(const std::string& name, int num) {
		customdefs[name] = num;
	}
	int operator()(const std::string& name) {
		// binary search
		int low = 0, high = numsysvars-1;
		while (low <= high) {
			int mid = (low+high) >> 1, cmp = name.compare(sysvars[mid].name);
			if (cmp < 0) {
				high = mid-1;
			} else if (cmp > 0) {
				low = mid+1;
			} else {
				return sysvars[mid].num;
			}
		}
		std::map<std::string, int>::const_iterator p = customdefs.find(name);
		if (p != customdefs.end()) {
			return p->second;
		}
		return -1;
	}
};

const std::string& sysvartostr(int num) {
	static const std::string emptystr = "";
	for (int i = numsysvars; i--;) {
		if (num == sysvars[i].num) return sysvars[i].name;
	}
	return emptystr;
}

// parses bot code and converts into bytecode
Code::Code(const std::string& filename) {
	std::ifstream file(filename.c_str(), std::ios::in);
	std::string token;
	long int temp;
	sysvarkeeper getsysvar;

	while (file >> token) {
		size_t i;
		if ((i = token.find('\'')) != std::string::npos) {
			std::string stemp;
			std::getline(file, stemp);
			if (i == 0) {
				if (tokens.empty()) {
					comment += token.substr(1);
					comment += stemp;
				}
				continue;
			}
			token.resize(i);
		}
		// make tokens lowercase for case-insensitive compare
		std::transform(token.begin(), token.end(), token.begin(), tolower);
		if (token == "def") {
			// definition
			std::string name;
			while (name.empty()) {
				if (!(file >> name)) return;
				size_t j;
				if ((j = name.find('\'')) != std::string::npos) {
					std::string stemp;
					std::getline(file, stemp);
					name.resize(i);
				}
			}
			std::string num;
			while (num.empty()) {
				if (!(file >> num)) return;
				size_t j;
				if ((j = num.find('\'')) != std::string::npos) {
					std::string stemp;
					std::getline(file, stemp);
					num.resize(i);
				}
			}
			int n = (int) strtoul(num.c_str(), NULL, 10);
			if (n > 32000) n = 32000;
			if (n < 0) n = 0;
			getsysvar.add_custom(name, n);
		} else if ((temp = getop(token)) != -1) {
			// operation
			tokens.push_back(OP);
			tokens.push_back(temp);
		} else if (token[0] == '*' && token.length() > 1) {
			// memory read
			if (token[1] == '.') {
				// memory read from sysvar
				if ((temp = getsysvar(token.substr(2))) != -1) {
					// valid sysvar
					tokens.push_back(READ);
					tokens.push_back(temp);
				}
			} else {
				temp = strtol(token.substr(1).c_str(), NULL, 10);
				// memory read from specific location
				if (temp > 32000) temp = 32000;
				if (temp <= 0) continue;
				tokens.push_back(READ);
				tokens.push_back(temp);
			}
		} else if (token[0] == '.') {
			// sysvar address
			if ((temp = getsysvar(token.substr(1))) != -1) {
				// valid sysvar
				tokens.push_back(NUM);
				tokens.push_back(temp);
			}
		} else {
			temp = strtol(token.c_str(), NULL, 10);
			// number
			if (temp > 32000) temp = 32000;
			if (temp < -32000) temp = -32000;
			tokens.push_back(NUM);
			tokens.push_back(temp);
		}
	}
	analyze();
}

void Code::analyze() {
	const int numtokens = tokens.size() >> 1;
	for (unsigned int i = 0; i < numtokens; ++i) {
		if (tokens.at(i*2) == READ) {
			if (tokens.at(i*2+1) > s_eye1 && tokens.at(i*2+1) <= s_eye9+1) ++myeye;
		}
		if (tokens.at(i*2) == NUM) {
			if (tokens.at(i*2+1) == s_shoot+1 && tokens.at(i*2+2) == OP) {
				int op = tokens.at(i*2+3);
				if (op == o_store || op == o_inc || op == o_dec) ++myshoot;
			}
		}
	}
}

void Code::code_exec(Bot& bot) const {
	const int stacksize = 20;
	int sp = 0, bsp = 0; // stack pointer, boolean stack pointer
	bool started = false, cond = false;
	long stack[stacksize];
	bool bstack[stacksize];

	const int numtokens = tokens.size() >> 1;

	for (unsigned int i = 0; i < numtokens; ++i) {
		switch (tokens.at(i*2)) {
			case NUM:
				if (sp < stacksize && (started || cond)) {
					stack[sp] = tokens.at(i*2+1);
					++sp;
				}
				break;
			case READ:
				if (sp < stacksize && (started || cond)) {
					stack[sp] = bot.data[tokens.at(i*2+1)-1];
					++sp;
				}
				break;
			case OP:
				int p = tokens.at(i*2+1);
				switch (p) {
					case o_store:
						if (sp > 1) {
							if (stack[sp-1] > 0 && stack[sp-1] <= 1000 && (bsp == 0 || bstack[bsp-1])) {
								if (stack[sp-2] > 32000) stack[sp-2] = 32000;
								if (stack[sp-2] < -32000) stack[sp-2] = -32000;
								bot.data[stack[sp-1]-1] = stack[sp-2];
								if (stack[sp-1]-1 == s_focuseye) bot.fill_refvars(); // refill focuseye - pretty bad solution but it works
								bot.nrg -= current_sim->prefs.storecost;
							}
							sp -= 2;
						}
						break;
					case o_inc:
						if (sp > 0) {
							if (stack[sp-1] > 0 && stack[sp-1] <= 1000 && (bsp == 0 || bstack[bsp-1])) {
								if (bot.data[stack[sp-1]-1] < 32000) ++bot.data[stack[sp-1]-1];
								if (stack[sp-1]-1 == s_focuseye) bot.fill_refvars(); // refill focuseye - pretty bad solution but it works
								bot.nrg -= current_sim->prefs.storecost*0.05; // reduced cost
							}
							--sp;
						}
						break;
					case o_dec:
						if (sp > 0) {
							if (stack[sp-1] > 0 && stack[sp-1] <= 1000 && (bsp == 0 || bstack[bsp-1])) {
								if (bot.data[stack[sp-1]-1] > -32000) --bot.data[stack[sp-1]-1];
								if (stack[sp-1]-1 == s_focuseye) bot.fill_refvars(); // refill focuseye - pretty bad solution but it works
								bot.nrg -= current_sim->prefs.storecost*0.05; // reduced cost
							}
							--sp;
						}
						break;
					case o_dupint:
						if (sp > 0 && sp < stacksize) {
							stack[sp] = stack[sp-1];
							++sp;
						}
						break;
					case o_dupbool:
						if (bsp > 0 && bsp < stacksize) {
							bstack[bsp] = stack[bsp-1];
							++bsp;
						}
						break;
					case o_incint:
						if (sp > 0) {
							++stack[sp-1];
						}
						break;
					case o_decint:
						if (sp > 0) {
							--stack[sp-1];
						}
						break;
					case o_neg:
						if (sp > 0) {
							stack[sp-1] = -stack[sp-1];
						}
						break;
					case o_dropbool:
						if (bsp > 0) {
							--bsp;
						}
						break;
					case o_swapint:
						if (sp > 1) {
							long temp = stack[sp-2];
							stack[sp-2] = stack[sp-1];
							stack[sp-1] = temp;
						}
						break;
					case o_swapbool:
						if (bsp > 1) {
							bool temp = bstack[bsp-2];
							bstack[bsp-2] = bstack[bsp-1];
							bstack[bsp-1] = temp;
						}
						break;
					case o_orint:
						if (sp > 1) {
							stack[sp-2] |= stack[sp-1];
							--sp;
						}
						break;
					case o_andint:
						if (sp > 1) {
							stack[sp-2] &= stack[sp-1];
							--sp;
						}
						break;
					case o_xorint:
						if (sp > 1) {
							stack[sp-2] ^= stack[sp-1];
							--sp;
						}
						break;
					case o_shr:
						if (sp > 1) {
							stack[sp-2] >>= stack[sp-1];
							--sp;
						}
						break;
					case o_shl:
						if (sp > 1) {
							stack[sp-2] <<= stack[sp-1];
							--sp;
						}
						break;
					case o_add:
						if (sp > 1) {
							long t1 = stack[--sp];
							stack[sp-1] += t1;
						}
						break;
					case o_sub:
						if (sp > 1) {
							long t1 = stack[--sp];
							stack[sp-1] -= t1;
						}
						break;
					case o_mult:
						if (sp > 1) {
							long t1 = stack[--sp];
							stack[sp-1] *= t1;
						}
						break;
					case o_div:
						if (sp > 1) {
							long t1 = stack[--sp];
							stack[sp-1] /= t1;
						}
						break;
					case o_mod:
						if (sp > 1) {
							long t1 = stack[--sp];
							stack[sp-1] %= t1;
						}
						break;
					case o_pow:
						if (sp > 1) {
							stack[sp-2] = int_pow(stack[sp-2], stack[sp-1]);
							--sp;
						}
						break;
					case o_read:
						if (sp > 0 && stack[sp-1] > 0 && stack[sp-1] <= 1000) {
							stack[sp-1] = bot.data[stack[sp-1]-1];
						}
						break;
					case o_start:
						if (cond) while (bsp > 1) {
							bstack[bsp-2] = bstack[bsp-2] && bstack[bsp-1];
							--bsp;
						}
						cond = false;
						started = true;
						break;
					case o_stop:
						cond = started = false;
						break;
					case o_pyth:
						if (sp > 1) {
							long t1 = stack[--sp];
							stack[sp-1] = pyth(stack[sp-1], t1);
						}
						break;
					case o_sqr:
						if (sp > 0) stack[sp-1] = static_cast<long>(sqrt(static_cast<float>(stack[sp-1])));
						break;
					case o_cond:
						cond = true;
						/* run through to clear boolean stack */
					case o_clearbool:
						bsp = 0;
						break;
					case o_c_lt:
						if (sp > 1 && bsp < stacksize) {
							bstack[bsp++] = stack[sp-2] < stack[sp-1];
							sp -= 2;
						}
						break;
					case o_c_gt:
						if (sp > 1 && bsp < stacksize) {
							bstack[bsp++] = stack[sp-2] > stack[sp-1];
							sp -= 2;
						}
						break;
					case o_c_eq:
						if (sp > 1 && bsp < stacksize) {
							bstack[bsp++] = stack[sp-2] == stack[sp-1];
							sp -= 2;
						}
						break;
					case o_c_le:
						if (sp > 1 && bsp < stacksize) {
							bstack[bsp++] = stack[sp-2] <= stack[sp-1];
							sp -= 2;
						}
						break;
					case o_c_ge:
						if (sp > 1 && bsp < stacksize) {
							bstack[bsp++] = stack[sp-2] >= stack[sp-1];
							sp -= 2;
						}
						break;
					case o_c_ne:
						if (sp > 1 && bsp < stacksize) {
							bstack[bsp++] = stack[sp-2] != stack[sp-1];
							sp -= 2;
						}
						break;
					case o_c_approxeq:
						if (sp > 1 && bsp < stacksize) {
							int tmp = stack[sp-2]/10;
							bstack[bsp++] = stack[sp-2]-tmp >= stack[sp-1] && stack[sp-2]+tmp <= stack[sp-1];
							sp -= 2;
						}
						break;
					case o_c_approxne:
						if (sp > 1 && bsp < stacksize) {
							int tmp = stack[sp-2]/10;
							bstack[bsp++] = stack[sp-2]-tmp < stack[sp-1] || stack[sp-2]+tmp > stack[sp-1];
							sp -= 2;
						}
						break;
					case o_c_true:
						if (bsp < stacksize) {
							bstack[bsp++] = true;
						}
						break;
					case o_c_false:
						if (bsp < stacksize) {
							bstack[bsp++] = false;
						}
						break;
					case o_c_and:
						if (bsp > 1) {
							bstack[bsp-2] = bstack[bsp-2] && bstack[bsp-1];
							--bsp;
						}
						break;
					case o_c_or:
						if (bsp > 1) {
							bstack[bsp-2] = bstack[bsp-2] || bstack[bsp-1];
							--bsp;
						}
						break;
					case o_c_xor:
						if (bsp > 1) {
							bstack[bsp-2] = (bool)(((int) bstack[bsp-2]) ^((int) bstack[bsp-1]));
							--bsp;
						}
						break;
					case o_c_not:
						if (bsp > 0) {
							bstack[bsp-1] = !bstack[bsp-1];
						}
						break;
					case o_dist:
						if (sp > 1) {
							const b2Vec2 pos = 100.0f*bot.b_body->GetPosition();
							float x = (float) stack[sp-2], y = (float) stack[sp-1];
							bot2field(x, y);
							stack[sp-2] = static_cast<long>(pyth(x-pos.x, y-pos.y));
							--sp;
						}
						break;
					case o_angle:
						if (sp > 1) {
							const b2Vec2 pos = 100.0f*bot.b_body->GetPosition();
							float x = (float) stack[sp-2], y = (float) stack[sp-1];
							bot2field(x, y);
							stack[sp-2] = 1256-rad2db(atan2(y-pos.y, x-pos.x));
							--sp;
						}
						break;
					case o_rnd:
						if (sp > 0) {
							if (stack[sp-1] != 0) {
								if (stack[sp-1] < 0) {
									stack[sp-1] = -bot.sim->rnd(0L, -stack[sp-1]);
								} else {
									stack[sp-1] = bot.sim->rnd(0L, stack[sp-1]);
								}
							}
						}
						break;
					case o_abs:
						if (sp > 0) {
							stack[sp-1] = abs(stack[sp-1]);
						}
						break;
					case o_sgn:
						if (sp > 0) {
							if (stack[sp-1] < 0) {
								stack[sp-1] = -1;
							} else if (stack[sp-1] > 0) {
								stack[sp-1] = 1;
							}
						}
						break;
					case o_nop:
						break;
					case o_overint:
						if (sp > 1 && sp < stacksize) {
							stack[sp] = stack[sp-2];
							++sp;
						}
						break;
					case o_overbool:
						if (bsp > 1 && bsp < stacksize) {
							bstack[bsp] = stack[bsp-2];
							++bsp;
						}
						break;
					//default:
						//std::cerr << "unimplemented bytecode:" << p << std::endl;
				}
				break;
		}
	}
}

// converts object to (an approximation of) the original bot file
std::string Code::toStr() const {
	const int numtokens = tokens.size() >> 1;
	std::ostringstream ostream;

	for (unsigned int i = 0; i < numtokens; ++i) {
		switch (tokens.at(i*2)) {
			case NUM:
				if (i != numtokens-1 && tokens.at(i*2+2) == OP) {
					// is a memory location
					int op = tokens.at(i*2+3);
					if (op == o_store || op == o_inc || op == o_dec) {
						const std::string& name = sysvartostr(tokens.at(i*2+1));
						if (name != "") {
							// sysvar name found
							ostream << '.' << name;
						} else {
							// no sysvar found
							ostream << tokens.at(i*2+1);
						}
					} else {
						ostream << tokens.at(i*2+1);
					}
				} else {
					ostream << tokens.at(i*2+1);
				}
				ostream << ' ';
				break;
			case READ: {
				ostream << '*';
				const std::string& name = sysvartostr(tokens.at(i*2+1));
				if (name != "") {
					ostream << '.' << name;
				} else {
					ostream << tokens.at(i*2+1);
				}
				ostream << ' ';
				break;
			}
			case OP:
				int p = tokens.at(i*2+1);
				for (unsigned int i = 0; i < num_ops; ++i) {
					if (ops[i].num == p) {
						ostream << ops[i].name << ' ';
						break;
					}
				}
				if (p == o_store || p == o_inc || p == o_dec || p == o_cond || p == o_start || p == o_stop || p == o_nop
				                || p == o_c_true || p == o_c_false || p == o_c_eq || p == o_c_ne || p == o_c_lt || p == o_c_gt
				                || p == o_c_le || p == o_c_ge || p == o_c_approxeq || p == o_c_approxne || p == o_c_and
				                || p == o_c_or || p == o_c_xor || p == o_c_not || p == o_dropbool || p == o_clearbool) {
					ostream << '\n';
				}
				if (p == o_stop) {
					// a second newline to separate genes
					ostream << '\n';
				}
				break;
		}
	}
	return ostream.str();
}
