#include "struct.h"

Shot::Shot(Sim& _sim, Bot& _parent, float x, float y) : age(0), parent(&_parent), sim(&_sim) {
	// create body
	b2BodyDef bdef;
	bdef.position.Set(x, y);
	bdef.userData = this;
	bdef.allowSleep = false;
	bdef.isSleeping = false;
	bdef.isBullet = false;
	bdef.fixedRotation = true;
	bdef.massData.I = 1.0e+15f;
	bdef.massData.center.Set(0.0f, 0.0f);
	bdef.massData.mass = 1.0e+15f; // very high to preserve velocity
	b_body = sim->b_world->CreateBody(&bdef);
	// create shape (very small)
	b2CircleDef sdef;
	sdef.friction = 0.0f;
	sdef.radius = 0.01f;
	sdef.filter.groupIndex = -4; // bots have +4, shots -4
	sdef.isSensor = true;
	b_body->CreateShape(&sdef);
	// add to parent shot's list
	parent->shots.push_back(*this);
	++sim->numshots;
}

void Shot::unlink() {
	if (parent != NULL) {
		parent->shots.erase(parent->shots.iterator_to(*this));
	} else {
		sim->orphan_shots.erase(sim->orphan_shots.iterator_to(*this));
	}
}

Shot::~Shot() {
	if (is_linked()) unlink();
	sim->b_world->DestroyBody(b_body);
	--sim->numshots;
}
