#include "ToolDrawer.h"

#include <wx/dc.h>
#include <wx/event.h>
#include <wx/image.h>

void ToolDrawer::CreateGrey() {
	greyimages.push_back(wxBitmap(images.back().ConvertToImage().ConvertToGreyscale()));
}
void ToolDrawer::Brighten(unsigned char& val) {
	#include "gamma_table.h"
	val = gamma[val];
}
void ToolDrawer::CreateLight() {
	wxImage image = images.back().ConvertToImage();
	if (!image.HasAlpha()) image.InitAlpha();
	unsigned char *data = image.GetData();
	for (int i = 24*24; i--; ) {
		Brighten(*data++);
		Brighten(*data++);
		Brighten(*data++);
	}
	lightimages.push_back(wxBitmap(image));
}
void ToolDrawer::AddItem(wxBitmap bmp, int eventid) {
	images.push_back(bmp);
	CreateGrey();
	CreateLight();
	eventids.push_back(eventid);
	issep.push_back(false);
}
void ToolDrawer::AddSeparator() {
	images.push_back(wxBitmap());
	greyimages.push_back(wxBitmap());
	lightimages.push_back(wxBitmap());
	issep.push_back(true);
}
int ToolDrawer::HitTest(int hx, int hy) const {
	if (hx < 2 || hx >= 26) return -1;
	int y = 2;
	for (int i = 0, j = 0; i < images.size(); ++i) {
		if (issep[i]) {
			y += 3;
		} else {
			if (hy >= y+2 && hy < y+26) return j;
			y += 28;
			++j;
		}
	}
	return -1;
}
bool ToolDrawer::DoActivate(int x, int y) {
	int item = HitTest(x, y);
	if (item != -1) {
		activated = item;
		return true;
	}
	activated = -1;
	return false;
}
bool ToolDrawer::DoHit(wxEvtHandler& handler, int x, int y, const bool disabled[]) const {
	int item = HitTest(x, y);
	if (item != -1 && !disabled[item]) {
		wxCommandEvent evt(wxEVT_COMMAND_TOOL_CLICKED, eventids[item]);
		handler.ProcessEvent(evt);
		return true;
	}
	return false;
}
void ToolDrawer::Draw(wxDC& dc, int x, const bool activate[], const bool disabled[]) const {
	x += 2;
	int y = 2;
	for (int i = 0, j = 0; i < images.size(); ++i) {
		if (issep[i]) {
			dc.DrawLine(x-1, y, x+25, y);
			y += 3;
		} else {
			if (disabled[j]) {
				dc.DrawBitmap(greyimages[i], x, y, true);
			} else if (activated == j || activate[j]) {
				dc.DrawBitmap(lightimages[i], x, y, true);
			} else {
				dc.DrawBitmap(images[i], x, y, true);
			}
			y += 28;
			++j;
		}
	}
}
