#include "redraw.h"

#include "struct.h"

#include <algorithm>

WxFieldDrawer::WxFieldDrawer(wxDC &_dc, int _viewx, int _viewy, int ww, int wh, unsigned int w, unsigned int h, float _zoom, bool _grey) :
		dc(_dc), nrg(*wxWHITE_PEN), body(_grey?wxColor(105, 105, 105):wxColor(255, 0, 255)),
		waste(_grey?wxColor(150, 150, 150):wxColor(0, 0, 255)),
		venom(_grey?wxColor(29, 29, 29):wxColor(0, 0, 255)),
		shell(_grey?wxColor(76, 76, 76):wxColor(255, 0, 0)),
		slime(_grey?wxColor(151, 151, 151):wxColor(255, 128, 0)),
		poison(_grey?wxColor(226, 226, 226):wxColor(255, 255, 0)),
		vtimer(_grey?wxColor(179, 179, 179):wxColor(0, 255, 255)),
		background(_grey?wxColor(15, 15, 15):wxColor(0, 0, 128)),
		width(w), height(h), w_width(ww), w_height(wh), viewx(_viewx), viewy(_viewy),
		scale(std::min(((float) ww)/((float) width), ((float) wh)/((float) height))), zoom(_zoom),
		atanzoom(atan(_zoom)*1.2), grey(_grey) {
	dc.SetBrush(background);
	dc.SetPen(*wxTRANSPARENT_PEN);
	dc.DrawRectangle(0, 0, (int) (((float) width)*zoom*scale), (int) (((float) height)*zoom*scale));
}

// convenience function for drawing the little coloured arcs inside bots
inline void BotArc(wxDC& dc, const wxPen& pen, int tx, int ty, float cosr, float sinr, float rot, float amount, float radius) {
	if (amount > 0.5f) {
		const float amt = amount/32000.0f*M_PI*2.0f+rot;
		const int ax = tx+(int) (cos(amt)*radius);
		const int ay = ty+(int) (sin(amt)*radius);
		const int bx = tx+(int) (cosr*radius);
		const int by = ty+(int) (sinr*radius);
		if ((ax != bx && ay != by) || amount > 16000.0f) {
			dc.SetPen(pen);
			dc.DrawArc(ax, ay, bx, by, tx, ty);
		}
	}
}

void WxFieldDrawer::DrawBot(const Bot &bot, bool bm) {
	b2Vec2 pos = bot.b_body->GetPosition();
	const int tx = (int) ((pos.x*100.0f-viewx)*zoom*scale+0.5f), ty = (int) ((pos.y*100.0f-viewy)*zoom*scale+0.5f);
	const float radius = bot.radius*zoom*scale;
	if (tx+radius < 0 || ty+radius < 0 || tx-radius > w_width || ty-radius > w_height) return;
	const unsigned char r = bot.color[0]*255.0f, g = bot.color[1]*255.0f, b = bot.color[2]*255.0f,
		gr = r*0.299f+g*0.587f+b*0.114f;
	if (radius > 4.0f) {
		dc.SetBrush(*wxTRANSPARENT_BRUSH);
		float rotation = bot.b_body->GetAngle();
		const float cosr = cos(rotation), sinr = sin(rotation);

		if (bm) {
			dc.SetPen(poison);
			dc.DrawCircle(tx, ty, static_cast<int>(radius+atanzoom*2+0.5));
			const float visarc = radius+(1440.0f*zoom*scale);
			dc.SetPen(nrg);
			dc.SetBrush(*wxTRANSPARENT_BRUSH);
			for (int i = 0; i < 9; ++i) {
				const float k = ((float) i)-4.0f;
				const float ten_deg = -0.17453292519943295769236907684886f,
					five_deg = -0.087266462599716478846184538424431f;
				const float cos_start = cos(rotation-k*ten_deg+five_deg),
					sin_start = sin(rotation-k*ten_deg+five_deg),
					cos_end = cos(rotation-k*ten_deg-five_deg),
					sin_end = sin(rotation-k*ten_deg-five_deg);
				dc.DrawLine(tx+cos_start*visarc, ty+sin_start*visarc, tx+cos_end*visarc, ty+sin_end*visarc/*, tx, ty*/);
				if (bot.eyes[i].eyeval && bot.eyes[i].id < current_sim->numbots) {
					const b2Vec2 pos2 = current_sim->bots.at(bot.eyes[i].id).b_body->GetPosition();
					const b2Vec2 dist = 100.0f*(pos-pos2);
					const float distf = pyth(dist.x, dist.y)-current_sim->bots.at(bot.eyes[i].id).radius;
					dc.SetPen(shell);
					dc.SetBrush(*wxWHITE_BRUSH);
					wxPoint points[4] = {wxPoint(cos_start*distf*scale*zoom, sin_start*distf*scale*zoom), wxPoint(cos_end*distf*scale*zoom, sin_end*distf*scale*zoom),
							wxPoint(cos_end*radius, sin_end*radius), wxPoint(cos_start*radius, sin_start*radius)};
					//dc.DrawArc(cos_start*distf, sin_start*distf, cos_end*distf, sin_end*distf, tx, ty);
					dc.DrawPolygon(4, points, tx, ty);
					//dc.SetPen(*wxBLACK_PEN);
					//dc.SetBrush(*wxTRANSPARENT_BRUSH);
					//dc.DrawLine(tx, ty, pos2.x*100.0f*scale*zoom, pos2.y*100.0f*scale*zoom);
				}
				dc.SetPen(nrg);
				dc.SetBrush(*wxTRANSPARENT_BRUSH);
				dc.DrawLine(tx+cos_start*radius, ty+sin_start*radius, tx+cos_start*visarc, ty+sin_start*visarc);
				dc.DrawLine(tx+cos_end*radius, ty+sin_end*radius, tx+cos_end*visarc, ty+sin_end*visarc);
			}
		}

		// draw outline
		wxPen botpen = wxPen(grey?wxColor(gr, gr, gr):wxColor(r, g, b));
		dc.SetPen(botpen);
		dc.DrawCircle(tx, ty, (int) radius);

		dc.DrawLine(tx, ty, tx+(int) (cosr*radius+0.5f), ty+(int) (sinr*radius+0.5f)); // draw aim

		#define ValArc(m, v) BotArc(dc, v, tx, ty, cosr, sinr, rotation, bot.v, m*radius);
		ValArc(0.875f, nrg);
		ValArc(0.75f, body);
		ValArc(0.625f, waste);
		ValArc(0.5f, venom);
		ValArc(0.375f, shell);
		ValArc(0.25f, slime);
		ValArc(0.125f, vtimer);
	} else {
		wxBrush botbrush(grey?wxColor(gr, gr, gr):wxColor(r, g, b));
		if (radius > 1.5f) {
			if (bm) {
				dc.SetPen(poison);
				dc.SetBrush(*wxTRANSPARENT_BRUSH);
				dc.DrawCircle(tx, ty, (int) (radius+atanzoom*2.0f+0.5f));
			}
			dc.SetBrush(botbrush);
			dc.SetPen(*wxTRANSPARENT_PEN);
			dc.DrawCircle(tx, ty, (int) (radius+0.5f));
		} else {
			if (bm) {
				dc.SetPen(poison);
				dc.SetBrush(*wxTRANSPARENT_BRUSH);
				dc.DrawCircle(tx, ty, 2+(int) (atanzoom*2.0f));
			}
			// minimum size 2x2 - now with delicious rectangular action
			dc.SetBrush(botbrush);
			dc.SetPen(*wxTRANSPARENT_PEN);
			dc.DrawRectangle(tx, ty, 2, 2);
		}
	}
}

void WxFieldDrawer::DrawShot(const Shot &shot) {
	const b2Vec2 pos = 100.0f*shot.b_body->GetPosition();
	const int tx = (int) (((pos.x-viewx)*scale*zoom)-atanzoom),
		ty = (int) (((pos.y-viewy)*scale*zoom)-atanzoom);
	if (tx < 0 || ty < 0 || tx > w_width || ty > w_height) return;
	const unsigned char r = shot.color[0]*255.0f, g = shot.color[1]*255.0f, b = shot.color[2]*255.0f,
		gr = r*0.299f+g*0.587f+b*0.114f;
	wxBrush brush(grey?wxColor(gr, gr, gr):wxColor(r, g, b));
	dc.SetBrush(brush);
	dc.SetPen(*wxTRANSPARENT_PEN);
	dc.DrawRectangle(tx, ty, (int) (atanzoom*2.0f), (int) (atanzoom*2.0f));
}

void WxFieldDrawer::DrawHit(const Hit &hit) {
	const int tx = ((int) (((float) (hit.x-viewx))*scale*zoom)), ty = ((int) (((float) (hit.y-viewy))*scale*zoom));
	if (tx < 0 || ty < 0 || tx > w_width || ty > w_height) return;
	const unsigned char r = hit.color[0]*255.0f, g = hit.color[1]*255.0f, b = hit.color[2]*255.0f,
		gr = r*0.299f+g*0.587f+b*0.114f;
	wxBrush brush(grey?wxColor(gr, gr, gr):wxColor(r, g, b));
	dc.SetBrush(brush);
	dc.SetPen(*wxTRANSPARENT_PEN);
	dc.DrawCircle(tx, ty, 4);
}

AGGFieldDrawer::AGGFieldDrawer(wxDC& _dc, wxImage& _back, int _viewx, int _viewy, int ww, int wh, int w, int h, float _zoom, bool _grey) :
		dc(_dc), backing(_back), width(w), height(h), w_width(ww), w_height(wh), viewx(_viewx), viewy(_viewy),
		scale(std::min(((float) ww)/((float) w), ((float) wh)/((float) h))),
		zoom(_zoom), atanzoom(atan(_zoom)*1.2), back_w(backing.GetWidth()), back_h(backing.GetHeight()),
		data(backing.GetData()), rbuffer(data, back_w, back_h, 3*back_w),
		pixf(rbuffer), renderer(pixf), grey(_grey) {
	// set background
	backing.SetRGB(wxRect(0, 0, back_w, back_h), 0, 0, 128);

	rast.reset();
	path.approximation_scale(10*zoom);

	// slowly scaling line width
	stroke.width(atanzoom);
}

#define AGGArc(radius, color, start, end) \
	path.init(tx, ty, radius, radius, start, end); \
	stroke.remove_all(); \
	path.rewind(0); \
	while ((pf = path.vertex(&x, &y)) != agg::path_cmd_stop) stroke.add_vertex(x, y, pf); \
	stroke.add_vertex(0, 0, agg::path_cmd_end_poly); \
	rast.reset(); \
	rast.add_path(stroke); \
	agg::render_scanlines_aa_solid(rast, scan, renderer, color);
#define AGGCircle(radius, color) AGGArc(radius, color, 0, M_PI*2);
#define AGGFilledArc(radius, color, start, end) \
	path.init(tx, ty, radius, radius, start, end); \
	rast.reset(); \
	rast.add_path(path); \
	agg::render_scanlines_aa_solid(rast, scan, renderer, color);
#define AGGFilledCircle(radius, color) AGGFilledArc(radius, color, 0, M_PI*2);
#define AGGLine(ax, ay, bx, by, color) \
	stroke.remove_all(); \
	stroke.add_vertex(ax, ay, agg::path_cmd_move_to); \
	stroke.add_vertex(bx, by, agg::path_cmd_line_to); \
	stroke.add_vertex(0, 0, agg::path_cmd_end_poly); \
	rast.reset(); \
	rast.add_path(stroke); \
	agg::render_scanlines_aa_solid(rast, scan, renderer, color);

void AGGFieldDrawer::DrawBot(const Bot &bot, bool bm) {
	const b2Vec2 pos = 100.0f*bot.b_body->GetPosition();
	if (pos.x < 0 || pos.x > ((float) width) || pos.y < 0 || pos.y > ((float) height)) {
		//std::cerr << "Warning: Bot " << i << " is out of the simulation, " << pos.x << "x" << pos.y << std::endl;
		return;
	}
	const float tx = (pos.x-viewx)*zoom*scale, ty = (pos.y-viewy)*zoom*scale,
		radius = std::max(bot.radius*zoom*scale, 1.0f);
	if (tx+radius < 0.0f || tx-radius > ((float) back_w) ||
		ty+radius < 0.0f || ty-radius > ((float) back_h)) return;
	unsigned int pf;
	double x, y;

	if (radius > 3.0f) {
		const double rotation = bot.b_body->GetAngle();
		if (bm) {
			AGGCircle(radius+atanzoom*2.0f, agg::rgba(1.0f, 1.0f, 0.0f));
			const float visarc = radius+(1440.0f*zoom*scale);
			const float ten_deg = -0.17453292519943295769236907684886f,
				five_deg = -0.087266462599716478846184538424431f;
			for (int i = 0; i < 9; ++i) {
				const float k = ((float) i)-4.0f;
				const float cos_start = cos(rotation-k*ten_deg+five_deg),
					sin_start = sin(rotation-k*ten_deg+five_deg),
					cos_end = cos(rotation-k*ten_deg-five_deg),
					sin_end = sin(rotation-k*ten_deg-five_deg);

				stroke.remove_all();
//				stroke.add_vertex(tx+cos_start*radius, ty+sin_start*radius, agg::path_cmd_move_to);
				stroke.add_vertex(tx+cos_start*visarc, ty+sin_start*visarc, agg::path_cmd_line_to);
				stroke.add_vertex(tx+cos_end*visarc, ty+sin_end*visarc, agg::path_cmd_line_to);
				stroke.add_vertex(tx+cos_end*radius, ty+sin_end*radius, agg::path_cmd_line_to);
				stroke.add_vertex(0.0f, 0.0f, 0);
				rast.reset();
				rast.add_path(stroke);
				agg::render_scanlines_aa_solid(rast, scan, renderer, agg::rgba(1.0f, 1.0f, 1.0f));
				//AGGLine(tx+cos_start*visarc, ty+sin_start*visarc, tx+cos_end*visarc, ty+sin_end*visarc);
				//AGGLine(tx+cos_start*radius, ty+sin_start*radius, tx+cos_start*visarc, ty+sin_start*visarc);
				//AGGLine(tx+cos_end*radius, ty+sin_end*radius, tx+cos_end*visarc, ty+sin_end*visarc);
				if (bot.eyes[i].eyeval && bot.eyes[i].id < current_sim->numbots) {
					const b2Vec2 pos2 = 100.0f*current_sim->bots.at(bot.eyes[i].id).b_body->GetPosition();
					const b2Vec2 dist = pos-pos2;
					const float distf = pyth(dist.x, dist.y)-current_sim->bots.at(bot.eyes[i].id).radius;
					const float x[4] = {cos_start*distf*scale*zoom, cos_end*distf*scale*zoom,
								cos_end*radius, cos_start*radius};
					const float y[4] = {sin_start*distf*scale*zoom, sin_end*distf*scale*zoom,
								sin_end*radius, sin_start*radius};
					// render inside of cell
					rast.reset();
					rast.add_vertex(x[0]+tx, y[0]+ty, agg::path_cmd_move_to);
					rast.add_vertex(x[1]+tx, y[1]+ty, agg::path_cmd_line_to);
					rast.add_vertex(x[2]+tx, y[2]+ty, agg::path_cmd_line_to);
					rast.add_vertex(x[3]+tx, y[3]+ty, agg::path_cmd_line_to);
					rast.add_vertex(0.0f, 0.0f, agg::path_cmd_end_poly);
					agg::render_scanlines_aa_solid(rast, scan, renderer, agg::rgba(1.0f, 1.0f, 1.0f));

					// render red border
					stroke.remove_all();
					stroke.add_vertex(x[0]+tx, y[0]+ty, agg::path_cmd_move_to);
					stroke.add_vertex(x[1]+tx, y[1]+ty, agg::path_cmd_line_to);
					stroke.add_vertex(x[2]+tx, y[2]+ty, agg::path_cmd_line_to);
					stroke.add_vertex(x[3]+tx, y[3]+ty, agg::path_cmd_line_to);
					stroke.add_vertex(0.0f, 0.0f, agg::path_cmd_end_poly);
					rast.reset();
					rast.add_path(stroke);
					agg::render_scanlines_aa_solid(rast, scan, renderer, agg::rgba(1.0f, 0.0f, 0.0f));
				}
			}
			const float cos_t = cos(rotation+4.5f*ten_deg),
				sin_t = sin(rotation+4.5f*ten_deg);
			stroke.remove_all();
			stroke.add_vertex(tx+cos_t*radius, ty+sin_t*radius, agg::path_cmd_move_to);
			stroke.add_vertex(tx+cos_t*visarc, ty+sin_t*visarc, agg::path_cmd_line_to);
			stroke.add_vertex(0.0f, 0.0f, 0);
			rast.reset();
			rast.add_path(stroke);
			agg::render_scanlines_aa_solid(rast, scan, renderer, agg::rgba(1.0f, 1.0f, 1.0f));
		}
		// draw outline
		AGGCircle(radius, agg::rgba(bot.color[0], bot.color[1], bot.color[2]));

		AGGLine(tx, ty, tx+cos(rotation)*radius, ty+sin(rotation)*radius, agg::rgba(bot.color[0], bot.color[1], bot.color[2])); // draw aim
		// convenience macro for drawing the little coloured arcs inside the bots
		#define AGGValArc(m, v, c) if (bot.v) { AGGArc(radius*(m), c, rotation, rotation+static_cast<double>(bot.v)*2*M_PI*0.00003125) }
		AGGValArc(0.875, nrg, agg::rgba(1.0, 1.0, 1.0));
		AGGValArc(0.75, body, agg::rgba(1.0, 0.0, 1.0));
		AGGValArc(0.625, waste, agg::rgba(0.0, 1.0, 0.0));
		AGGValArc(0.5, venom, agg::rgba(0.0, 0.0, 1.0));
		AGGValArc(0.375, shell, agg::rgba(1.0, 0.0, 0.0));
		AGGValArc(0.25, slime, agg::rgba(1.0, 0.5, 0.0));
		AGGValArc(0.125, vtimer, agg::rgba(0.0, 1.0, 1.0));
	} else {
		if (bm) {
			AGGCircle(radius+atanzoom*2, agg::rgba(1.0, 1.0, 0.0));
		}
		AGGFilledCircle(radius, agg::rgba(bot.color[0], bot.color[1], bot.color[2]));
	}
}

// alpha blending function
inline void blend(unsigned char* ptr, float alpha, float r, float g, float b) {
	alpha = sqrt(alpha); // increase gamma a little
	*ptr *= 1.0f-alpha;
	*ptr += alpha*((unsigned char) (r*alpha*255.0f));
	*(ptr+1) *= 1.0f-alpha;
	*(ptr+1) += alpha*((unsigned char) (g*alpha*255.0f));
	*(ptr+2) *= 1.0f-alpha;
	*(ptr+2) += alpha*((unsigned char) (b*alpha*255.0f));
}

void AGGFieldDrawer::DrawShot(const Shot& shot) {
	const b2Vec2 pos = 100.0f*shot.b_body->GetPosition();
	if (pos.x < -slop || pos.x > ((float) width)+slop || pos.y < -slop || pos.y > ((float) height)+slop) {
//		std::cerr << "Warning: Shot " << i << " is out of the simulation, " << pos.x << "x" << pos.y << std::endl;
		return;
	}
	const float tx = (pos.x-viewx)*zoom*scale, ty = (pos.y-viewy)*zoom*scale;
	const int x = (int) tx, y = (int) ty;
	if (tx < 0.0f || tx >= (float) back_w || ty < 0.0f || ty >= (float) back_h) return;

	unsigned char* line = rbuffer.row_ptr(y)+(x*3);
	// antialiased, subpixel-precision point
	blend(line, (1.0f-(ty-(float) y))*(1.0f-(tx-(float) x)), shot.color[0], shot.color[1], shot.color[2]);
	blend(line+3, (1.0f-(ty-(float) y))*(tx-(float) x), shot.color[0], shot.color[1], shot.color[2]);
	line = rbuffer.row_ptr(y+1)+x*3;
	blend(line, (ty-(float) y)*(1.0f-(tx-(float) x)), shot.color[0], shot.color[1], shot.color[2]);
	blend(line+3, (ty-(float) y)*(tx-(float) x), shot.color[0], shot.color[1], shot.color[2]);
}

void AGGFieldDrawer::DrawHit(const Hit& hit) {
	if (hit.x > width || hit.y > height) {
		return;
	}
	const float tx = ((float) (hit.x-viewx))*zoom*scale, ty = ((float) (hit.y-viewy))*zoom*scale;
	if (tx < 0.0f || tx >= (float) back_w || ty < 0.0f || ty >= (float) back_h) {
		return;
	}

	AGGFilledCircle(3.0f, agg::rgba(hit.color[0], hit.color[1], hit.color[2]));
}

AGGFieldDrawer::~AGGFieldDrawer() {
	wxBitmap bitmap(grey?backing.ConvertToGreyscale():backing);
	dc.DrawBitmap(bitmap, 0, 0, false);
}
