#ifndef REDRAW_H
#define REDRAW_H

#include <wx/dc.h>
#include <wx/image.h>

#include "agg_basics.h"
#include "agg_rendering_buffer.h"
#include "agg_rasterizer_scanline_aa.h"
#include "agg_arc.h"
#include "agg_scanline_u.h"
#include "agg_renderer_scanline.h"
#include "agg_pixfmt_rgb.h"
#include "agg_path_storage.h"
#include "agg_vcgen_stroke.h"

#include <algorithm>

class Bot;
struct Shot;
struct Hit;

class FieldDrawer {
public:
	virtual ~FieldDrawer() { }
	virtual void DrawBot(const Bot& bot, bool bm) = 0;
	virtual void DrawShot(const Shot& shot) = 0;
	virtual void DrawHit(const Hit& hit) = 0;
};

class WxFieldDrawer {
public:
	WxFieldDrawer(wxDC& dc, int viewx, int viewy, int ww, int wh, unsigned int w, unsigned int h, float zoom, bool grey);
	~WxFieldDrawer() { }
	void DrawBot(const Bot& bot, bool bm);
	void DrawShot(const Shot& shot);
	void DrawHit(const Hit& hit);
private:

	wxDC& dc;
	const wxPen nrg, body, waste, venom, shell, slime, poison, vtimer;
	const wxBrush background;

	const int width, height, w_width, w_height, viewx, viewy;
	const float scale, zoom, atanzoom;
	const bool grey;
};

class AGGFieldDrawer {
public:
	AGGFieldDrawer(wxDC& dc, wxImage& back, int viewx, int viewy, int ww, int wh, int w, int h, float zoom, bool grey);
	~AGGFieldDrawer();
	void DrawBot(const Bot& bot, bool bm);
	void DrawShot(const Shot& shot);
	void DrawHit(const Hit& hit);
	static void BackingSize(int ww, int wh, int w, int h, float zoom, int& back_w, int& back_h) {
		float scale = std::min(((float) ww)/((float) w), ((float) wh)/((float) h));
		back_w = std::min((int) (((float) w)*scale*zoom), ww);
		back_h = std::min((int) (((float) h)*scale*zoom), wh);
	}
private:
	wxDC& dc;
	wxImage& backing;
	const unsigned int width, height, w_width, w_height;
	const int viewx, viewy;
	const float scale, zoom, atanzoom;
	const int back_w, back_h;
	unsigned char* data;
	agg::rendering_buffer rbuffer;
	agg::pixfmt_rgb24 pixf;
	agg::renderer_base<agg::pixfmt_rgb24> renderer;
	agg::scanline_u8 scan;
	agg::rasterizer_scanline_aa<> rast;
	agg::arc path;
	agg::vcgen_stroke stroke;
	const bool grey;
};

#endif
