#ifndef STRUCT_H
#define STRUCT_H

class wxProgressDialog;

#include <vector>
#include <list>
#include <boost/intrusive/list.hpp>
#include "ptr_vector.h"
#include <string>
#include <iostream>

#include "Box2D.h"

#include <boost/serialization/array.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/list.hpp>
#include <boost/serialization/shared_ptr.hpp>
#include <boost/serialization/nvp.hpp>

#include <boost/shared_ptr.hpp>
#include <boost/pool/pool.hpp>

/* *** enums *** */

// types of code bytecodes
enum {
	END,
	NUM,
	OP,
	READ
};

// veggy feeding types
enum {
	NRG_PER_CYCLE,
	NRG_PER_KILOBODY
};

// choice of renderers
enum {
	RENDERER_WX,
	RENDERER_AGG
};

// objects contained in a sim
/*enum ObjectType {
	BOT,
	SHOT
};*/

/* *** class declarations *** */

// keeps track of venom effects, etc.
struct VenomData {
	int loc, val;
	unsigned int time;
	bool poison;
	template<class Archive> void serialize(Archive& ar, const unsigned int version) {
		ar & BOOST_SERIALIZATION_NVP(loc);
		ar & BOOST_SERIALIZATION_NVP(val);
		ar & BOOST_SERIALIZATION_NVP(time);
	}
};

class Bot;

// bytecode representation of a DNA file
struct Code {
	typedef boost::shared_ptr<Code> ptr;
	std::vector<int> tokens;
	int myeye, myshoot;

	Code() { }
	Code(const std::string& filename);
	void analyze();
	void code_exec(Bot& bot) const;
/*	wxString toStr() const;
	static Code::ptr fromStr(const wxString& str);*/
	template<class Archive> void serialize(Archive& ar, const unsigned int version) {
		ar & BOOST_SERIALIZATION_NVP(tokens);
		ar & BOOST_SERIALIZATION_NVP(myeye);
		ar & BOOST_SERIALIZATION_NVP(myshoot);
	}
};

// "template" for creating a bot
struct BotTemplate {
	std::string name; // bot name and filename
	Code::ptr code;
	float nrg, body, color[3]; // starting nrg and body, bot colour
	unsigned int numbots; // number of bots to create
	bool veg, fixed, noexec, novis; // some flags
	template<class Archive> void serialize(Archive& ar, const unsigned int version) {
		ar & BOOST_SERIALIZATION_NVP(name);
		ar & BOOST_SERIALIZATION_NVP(code);
		ar & BOOST_SERIALIZATION_NVP(nrg);
		ar & BOOST_SERIALIZATION_NVP(body);
		ar & BOOST_SERIALIZATION_NVP(color);
		ar & BOOST_SERIALIZATION_NVP(numbots);
		ar & BOOST_SERIALIZATION_NVP(veg);
		ar & BOOST_SERIALIZATION_NVP(fixed);
		ar & BOOST_SERIALIZATION_NVP(noexec);
		ar & BOOST_SERIALIZATION_NVP(novis);
	}
};

/*// very simple base class for objects in a sim (bots and shots)
struct SimObject {
	ObjectType type; // to keep track of the type of object
	SimObject(ObjectType _t) : type(_t) { }
};*/

struct EyeInfo {
	size_t id;
	int eyeval;
	int refxpos, refypos, refnrg, refbody, refshell, refpoison, refvenom, refvel,
		refvelup, refveldn, refveldx, refvelsx, refvelscalar, refeye, refshoot,
		in1, in2, in3, in4, in5, in6, in7, in8, in9, in10, memval;
	template<class Archive> void serialize(Archive& ar, const unsigned int version) {
		ar & BOOST_SERIALIZATION_NVP(id)
		   & BOOST_SERIALIZATION_NVP(eyeval)
		   & BOOST_SERIALIZATION_NVP(refxpos)
		   & BOOST_SERIALIZATION_NVP(refypos)
		   & BOOST_SERIALIZATION_NVP(refnrg)
		   & BOOST_SERIALIZATION_NVP(refbody)
		   & BOOST_SERIALIZATION_NVP(refshell)
		   & BOOST_SERIALIZATION_NVP(refpoison)
		   & BOOST_SERIALIZATION_NVP(refvenom)
		   & BOOST_SERIALIZATION_NVP(refvel)
		   & BOOST_SERIALIZATION_NVP(refvelup)
		   & BOOST_SERIALIZATION_NVP(refveldn)
		   & BOOST_SERIALIZATION_NVP(refveldx)
		   & BOOST_SERIALIZATION_NVP(refvelsx)
		   & BOOST_SERIALIZATION_NVP(refeye)
		   & BOOST_SERIALIZATION_NVP(refshoot)
		   & BOOST_SERIALIZATION_NVP(in1)
		   & BOOST_SERIALIZATION_NVP(in2)
		   & BOOST_SERIALIZATION_NVP(in3)
		   & BOOST_SERIALIZATION_NVP(in4)
		   & BOOST_SERIALIZATION_NVP(in5)
		   & BOOST_SERIALIZATION_NVP(in6)
		   & BOOST_SERIALIZATION_NVP(in7)
		   & BOOST_SERIALIZATION_NVP(in8)
		   & BOOST_SERIALIZATION_NVP(in9)
		   & BOOST_SERIALIZATION_NVP(in10)
		   & BOOST_SERIALIZATION_NVP(memval);
	}
};

class Sim;
class Bot;

// a shot, fired by a bot
//typedef boost::intrusive::list_base_hook<boost::intrusive::link_mode<boost::intrusive::auto_unlink> > auto_unlink_hook;
typedef boost::intrusive::list_base_hook<boost::intrusive::link_mode<boost::intrusive::safe_link> > safe_link_hook;
class Shot : public safe_link_hook {
public:
	// most of this is from the original Darwinbots source and doesn't seem to make much sense
	// TODO: reimplement
	unsigned long nrg, age;
	float body, range, val, color[3];
	int shoot, loc, vval;
	//size_t id;
	//bool parent_exists; // marks if parent is still alive
	Bot* parent;
	Sim* sim;
	b2Body* b_body; // Box2D body

	Shot() : parent(NULL), sim(NULL) { }
	Shot(Sim& sim, Bot& parent, float x, float y);
	~Shot();

	void reparent(Bot& newparent);
//	void unlink() { auto_unlink_hook::unlink(); }
	void unlink();
//	bool is_linked() { return auto_unlink_hook::is_linked(); }

	template<class Archive> void serialize(Archive& ar, const unsigned int version) {
		//ar & BOOST_SERIALIZATION_NVP(b_body);
		ar & BOOST_SERIALIZATION_NVP(nrg)
		   & BOOST_SERIALIZATION_NVP(age)
		   & BOOST_SERIALIZATION_NVP(body)
		   & BOOST_SERIALIZATION_NVP(range)
		   & BOOST_SERIALIZATION_NVP(val)
		   & BOOST_SERIALIZATION_NVP(color)
		   & BOOST_SERIALIZATION_NVP(shoot)
		   & BOOST_SERIALIZATION_NVP(loc)
		   & BOOST_SERIALIZATION_NVP(vval);
		//   & BOOST_SERIALIZATION_NVP(parent);
		//ar & BOOST_SERIALIZATION_NVP(parent_exists);
		// ar & BOOST_SERIALIZATION_NVP(id);
	}
};

typedef boost::intrusive::list<Shot, boost::intrusive::constant_time_size<true> > shotlist;

// a bot
class Bot {
public:
	//typedef boost::shared_ptr<Bot> ptr; // typedef'd smart pointer to a Bot
	unsigned int robage; // age of the robot
	std::string name; // bot name
	Code::ptr code; // bot bytecode
	Sim* sim; // tracks the owning sim
	std::list<VenomData> venomed; // list of venom locations/values/durations
	//std::set<size_t> shots; // list of IDs of shots shot by this bot
	shotlist shots; // shots fired by this bot
	float nrg, body, waste, shell, slime, poison, venom, vtimer, radius; // bot properties
	float color[3]; // bot colour
	bool veg, fixed, noexec, novis; // flags
	bool radius_changed; // flags if the bot's size has changed and its shape should be recreated
//	bool dead; // flags if the bot has been killed and should be deleted
	int kills; // number of kills made by this bot
	short data[1000]; // bot memory
	//unsigned long eye[9]; // identifiers of bots seen in the bot's eyes
	//int eyeval[9];
	EyeInfo eyes[9];

	b2Body* b_body; // Box2D body
	b2Shape* b_shape; // Box2D shape
	size_t id; // this bot's identifier

	//void fill_eye(unsigned long id, int eye, int eyeval);
	bool docode();
	void create_substances();
	void move();
	void repro();
	void see(const Bot& b); // sight check - this bot and 'b'
	void fill_refvars(); // sets the "ref*" variables in bot memory based on seen bot
	void apply_venom(int loc, int val, int shootval, bool poison); // poisons the bot
	bool hit(Shot& shot); // determines effects of being hit by a shot
	void shoot(); // shoots a shot
	void vision();
	float Mass() const; // calculates this bot's mass
	void ResetMass(); // sets this bot's mass properties for Box2D

	//ObjectType type() const { return BOT; }

	template<class Archive> void serialize(Archive& ar, const unsigned int version) {
		ar & BOOST_SERIALIZATION_NVP(robage)
		   & BOOST_SERIALIZATION_NVP(name)
		   & BOOST_SERIALIZATION_NVP(code)
		//   & BOOST_SERIALIZATION_NVP(sim)
		   & BOOST_SERIALIZATION_NVP(venomed)
		//   & BOOST_SERIALIZATION_NVP(shots)
		   & BOOST_SERIALIZATION_NVP(nrg)
		   & BOOST_SERIALIZATION_NVP(body)
		   & BOOST_SERIALIZATION_NVP(waste)
		   & BOOST_SERIALIZATION_NVP(shell)
		   & BOOST_SERIALIZATION_NVP(slime)
		   & BOOST_SERIALIZATION_NVP(poison)
		   & BOOST_SERIALIZATION_NVP(venom)
		   & BOOST_SERIALIZATION_NVP(vtimer)
		   & BOOST_SERIALIZATION_NVP(radius)
		   & BOOST_SERIALIZATION_NVP(color)
		   & BOOST_SERIALIZATION_NVP(veg)
		   & BOOST_SERIALIZATION_NVP(fixed)
		   & BOOST_SERIALIZATION_NVP(noexec)
		   & BOOST_SERIALIZATION_NVP(novis)
		   & BOOST_SERIALIZATION_NVP(radius_changed)
		   & BOOST_SERIALIZATION_NVP(kills)
		   & BOOST_SERIALIZATION_NVP(data)
		   & BOOST_SERIALIZATION_NVP(eyes);
		//   & BOOST_SERIALIZATION_NVP(b_body)
		//   & BOOST_SERIALIZATION_NVP(b_shape)
		//   & BOOST_SERIALIZATION_NVP(id)
	}
	// ordinary constructor
	Bot(Sim* parent, float x, float y, float ang, float body, Code::ptr code);
	// default constructor
	Bot() { }
	~Bot();
};

// purely aesthetic hit where a shot hits a bot
struct Hit {
	unsigned int x, y; // location of the hit
	float color[3]; // colour
	char life; // lifetime of hit, starts at 4 and decrements each cycle
};

// listens for shot-bot collisions
class ShotListener : public b2ContactListener {
public:
	void Add(const b2ContactPoint* point);
	void Persist(const b2ContactPoint* point) { };
	void Remove(const b2ContactPoint* point) { };
};

// stores settings
struct SimSettings {
	float movcost, shotcost, bodyupkeepcost, poisoncost, venomcost, shellcost,
		slimecost, friction, move_eff, elast, storecost, agecost, botdensity;
	unsigned long vegfeed, vegfeed_type,width, height, maxnonveg, maxveg,
		vegrepop_threshold, vegrepop_amount, vegrepop_cooldown, maxvel;
	std::vector<BotTemplate> templates;
	SimSettings();
	void Save();
	template<class Archive> void serialize(Archive & ar, const unsigned int version) {
		ar & BOOST_SERIALIZATION_NVP(movcost)		& BOOST_SERIALIZATION_NVP(shotcost)
		   & BOOST_SERIALIZATION_NVP(bodyupkeepcost)	& BOOST_SERIALIZATION_NVP(poisoncost)
		   & BOOST_SERIALIZATION_NVP(venomcost)		& BOOST_SERIALIZATION_NVP(shellcost)
		   & BOOST_SERIALIZATION_NVP(slimecost)		& BOOST_SERIALIZATION_NVP(friction)
		   & BOOST_SERIALIZATION_NVP(move_eff)		& BOOST_SERIALIZATION_NVP(elast)
		   & BOOST_SERIALIZATION_NVP(storecost)		& BOOST_SERIALIZATION_NVP(agecost)
		   & BOOST_SERIALIZATION_NVP(botdensity)	& BOOST_SERIALIZATION_NVP(vegfeed)
		   & BOOST_SERIALIZATION_NVP(vegfeed_type)	& BOOST_SERIALIZATION_NVP(width)
		   & BOOST_SERIALIZATION_NVP(height)		& BOOST_SERIALIZATION_NVP(maxnonveg)
		   & BOOST_SERIALIZATION_NVP(maxveg)		& BOOST_SERIALIZATION_NVP(vegrepop_threshold)
		   & BOOST_SERIALIZATION_NVP(vegrepop_amount)	& BOOST_SERIALIZATION_NVP(vegrepop_cooldown)
		   & BOOST_SERIALIZATION_NVP(maxvel)		& BOOST_SERIALIZATION_NVP(templates);
	}
};

// keeps track of a simulation
class Sim {
public:
	typedef boost::shared_ptr<Sim> ptr; // smart pointer

	b2World* b_world; // the Box2D world associated with the simulation
	b2MouseJoint* mjoint; // the mouse joint for moving bots around
	size_t numbots, numshots, numveg;
	unsigned long cycles, bmx, bmy, bmid, lastrepop;
	bool bm, down; // "is a bot selected?" and "is the mouse being held down on a bot?"
	ShotListener listen; // above bot-shot collision listener

	boost::pool<> shotpool;

	// all bots, hits, plus orphaned shots
	ptr_vector<Bot> bots;
	shotlist orphan_shots;
	std::list<Hit> hits;

	SimSettings prefs; // sim settings

	Sim() : lastrepop(0), bm(false), down(false), shotpool(sizeof(Shot)) { }
	Sim(SimSettings &_prefs);
	~Sim();
	void populate(const BotTemplate &t); // adds bots to the sim based on its template
	size_t add_bot(float x, float y, float ang, float body, Code::ptr code); // adds one bot to the sim
	void del_bot(size_t id); // kills a bot
	Shot& add_shot(Bot& parent, float x, float y); // adds a shot to the sim
	void del_shot(Shot&); // destroys a shot
	bool cycle(); // calculates one cycle

	class cancelled { } cancelled_exception;

	template<class Archive> void save(Archive& ar, const unsigned int version, wxProgressDialog& pdialog) const;
	template<class Archive> void load(Archive& ar, const unsigned int version, wxProgressDialog& pdialog);

//	BOOST_SERIALIZATION_SPLIT_MEMBER()
};

extern Sim* current_sim;

/* *** generic mathy routines *** */

// floating point and integer Pythagorean theorem
inline float pyth(float a, float b) {
	return sqrt(a*a+b*b);
}
inline int pyth(int a, int b) {
	float x = static_cast<float>(a), y = static_cast<float>(b);
	return static_cast<int>(sqrt(x*x+y*y));
}
inline long pyth(long a, long b) {
	float x = static_cast<float>(a), y = static_cast<float>(b);
	return static_cast<long>(sqrt(x*x+y*y));
}

// rotates a vector by a certain angle
inline float vec_rotate(b2Vec2& vec, float amount) {
	const float ang = atan2(vec.y, vec.x)+amount;
	const float mag = sqrt((vec.x*vec.x)+(vec.y*vec.y));
	vec.x = cos(ang)*mag;
	vec.y = sin(ang)*mag;
	return mag;
}

// converts between DB units and radians (1256 == 2pi)
inline int rad2db(float rad) {
	return static_cast<int>(rad*199.8986085234206);
}
inline float db2rad(int db) {
	return static_cast<float>(db)*0.0050025360725952;
}

// converts between "bot's coordinates" and "field coordinates" when field is larger than 32000x32000
// modifies parameters in place
inline void bot2field(float& botx, float& boty) {
	const unsigned int width = current_sim->prefs.width, height = current_sim->prefs.height;
	if (width <= 32000 || height <= 32000) return; // null conversion
	//if (botx < 0 || botx >= 32000 || boty < 0 || boty >= 32000) return; // out of bounds
	// ratios
	const float x = (float) width/32000.0f, y = (float) height/32000.0f;
	botx *= x;
	boty *= y;
}
inline void field2bot(float& botx, float& boty) {
	const unsigned int width = current_sim->prefs.width, height = current_sim->prefs.height;
	if (width <= 32000 || height <= 32000) return; // null conversion
	//if (botx < 0 || botx >= width || boty < 0 || boty >= height) return; // out of bounds
	// ratios
	const float x = 32000.0f/(float) width, y = 32000.0f/(float) height;
	botx *= x;
	boty *= y;
}
inline void bot2field(int& botx, int& boty) {
	float x = (float) botx, y = (float) boty;
	bot2field(x, y);
	botx = round(x);
	boty = round(y);
}
inline void field2bot(int& botx, int& boty) {
	float x = (float) botx, y = (float) boty;
	field2bot(x, y);
	botx = round(x);
	boty = round(y);
}

/*
// generic "round integer" routine
inline int round(float x) {
	return (int) x+0.5f;
}
*/

// eye width to float
inline float width_to_range(float width) {
	return 1440.0f*(1.0f-(log(width/0.17453292519943295769f)*0.25f));
}

// converts a body value into a radius
inline float getradius(float body) {
	// as simplified from DB source code
	const float magic_body = 1.004618978534739692776156516629271209239959716796875;
		// the body at which radius becomes 1 (quite accurate)
	if (body < magic_body) body = magic_body;
	return pow(log(body)*body, 0.3333333333333333333333333f)*6.0f;
}

// integer-integer exponentiation
inline long int_pow(long base, long exp) {
	long r = 1;
	while (exp) {
		if (exp & 1) r *= base;
		base *= base;
		exp >>= 1;
	}
	return r;
}

// random numbers
inline float randf(float min, float max) {
	return ((float) rand())*(max-min)/((float) RAND_MAX)+min;
}

inline int randint(int min, int max) {
	if (min == max) return min;
	return rand()%(max-min+1)+min;
}

/* *** inlined class methods *** */

inline float Bot::Mass() const {
	const float r = body*0.001f+shell*0.005f;
	if (r < 1.0f) return 1.0f;
	return r;
}

inline void Bot::ResetMass() {
	b2MassData mass;
	mass.mass = Mass()*0.3f;
	if (fixed) mass.mass *= 10000.0f;
	mass.center.SetZero();
	//mass.I = tmpmass*100.0f;
	mass.I = 0.0f;
	b_body->SetMass(&mass);
}

inline Shot& Sim::add_shot(Bot& parent, float x, float y) {
	Shot* shot = (Shot*) shotpool.malloc();
	new(shot) Shot(*this, parent, x, y);
	return *shot;
}

inline void Sim::del_shot(Shot& shot) {
	shot.~Shot();
	shotpool.free(&shot);
}

inline void Shot::reparent(Bot& newparent) {
	unlink();
	newparent.shots.push_back(*this);
	parent = &newparent;
}

// threshold for declaring things "out of the sim", in Box2d units
const float slop = 0.005f;

#endif
