#include <wx/wxprec.h>

#ifdef __BORLANDC__
	#pragma hdrstop
#endif

#ifndef WX_PRECOMP
	#include <wx/wx.h>
#endif

#include <wx/config.h>
#include <wx/dcbuffer.h>
#include <wx/mstream.h>
#include <wx/image.h>
#include <wx/filedlg.h>
#include <wx/progdlg.h>
#include <wx/aboutdlg.h>

#include "Box2D.h"

#include <vector>
#include <algorithm>
#include <fstream>

#include <boost/archive/xml_oarchive.hpp>
#include <boost/archive/xml_iarchive.hpp>
#include <boost/iostreams/filtering_stream.hpp>
#include <boost/iostreams/filtering_streambuf.hpp>
#include <boost/iostreams/copy.hpp>
#include <boost/iostreams/filter/bzip2.hpp>

#include "wxBots.h"
#include "struct.h"
#include "redraw.h"
#include "Sim_serialization.h"

#include "OptionsDialog.h"
#include "NewDialog.h"
#include "BotProperties.h"

#include "icons.h"

#if defined(__WXGTK__) || defined(__WXMOTIF__)
#include "icon/icon_16.xpm"
#include "icon/icon_24.xpm"
#include "icon/icon_32.xpm"
#include "icon/icon_48.xpm"
#endif

BEGIN_EVENT_TABLE(MainFrame, wxFrame)
	EVT_MENU(-1, MainFrame::OnEvent)
	EVT_CLOSE(MainFrame::OnClose)
	EVT_IDLE(MainFrame::OnIdle)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(wxBotView, wxWindow)
	EVT_PAINT(wxBotView::OnPaint)
	EVT_LEFT_DOWN(wxBotView::OnDown)
	EVT_LEFT_UP(wxBotView::OnUp)
	EVT_LEFT_DCLICK(wxBotView::OnDClick)
	EVT_RIGHT_UP(wxBotView::OnRUp)
	EVT_MOTION(wxBotView::OnMove)
	EVT_MOUSEWHEEL(wxBotView::OnWheel)
	EVT_MOUSE_CAPTURE_LOST(wxBotView::MouseCaptureLost)
END_EVENT_TABLE()

IMPLEMENT_APP(App)

wxBitmap* bmp_new, * bmp_load, * bmp_save, * bmp_play, * bmp_pause, * bmp_cycle,
		* bmp_nodisplay, * bmp_zoomin, * bmp_zoomout, * bmp_zoomnone;

Sim* current_sim = NULL;

bool cycle;

inline void cutimg(const wxImage& img, wxBitmap*& bmp) {
	static int x = 0;
	bmp = new wxBitmap(img.GetSubImage(wxRect(x, 0, 24, 24)));
	x += 24;
}

bool App::OnInit() {
	if (!wxApp::OnInit())
		return false;

	SetAppName(_T("wxBots"));

	// load toolbar images
	wxImage::AddHandler(new wxPNGHandler);
	wxMemoryInputStream stream(icons, sizeof(icons));
	const wxImage img_icons(stream, wxBITMAP_TYPE_PNG);
#define mkimg(bmp_out) cutimg(img_icons, bmp_out)
	mkimg(bmp_new);
	mkimg(bmp_load);
	mkimg(bmp_save);
	mkimg(bmp_play);
	mkimg(bmp_pause);
	mkimg(bmp_cycle);
	mkimg(bmp_nodisplay);
	mkimg(bmp_zoomin);
	mkimg(bmp_zoomout);
	mkimg(bmp_zoomnone);
#undef mkimg

	wxConfigBase* cfg = wxConfigBase::Get();
	window = new MainFrame(_T("wxBots"), wxPoint(cfg->Read(_T("/Window/X"), -1), cfg->Read(_T("/Window/Y"), -1)),
		wxSize(cfg->Read(_T("Window/Width"), 800), cfg->Read(_T("Window/Height"), 600))); // TODO: write the corresponding save position code
	wxIconBundle icons;
#ifdef __WXMSW__
	{
		wxIcon tmp(_T("mainicon"), wxBITMAP_TYPE_ICO_RESOURCE, 16, 16);
		icons.AddIcon(tmp);
	}
	{
		wxIcon tmp(_T("mainicon"), wxBITMAP_TYPE_ICO_RESOURCE, 24, 24);
		icons.AddIcon(tmp);
	}
	{
		wxIcon tmp(_T("mainicon"), wxBITMAP_TYPE_ICO_RESOURCE, 32, 32);
		icons.AddIcon(tmp);
	}
	{
		wxIcon tmp(_T("mainicon"), wxBITMAP_TYPE_ICO_RESOURCE, 48, 48);
		icons.AddIcon(tmp);
	}
#endif
#if defined(__WXGTK__) || defined(__WXMOTIF__)
	{
		wxIcon tmp(icon_16_xpm);
		icons.AddIcon(tmp);
	}
	{
		wxIcon tmp(icon_24_xpm);
		icons.AddIcon(tmp);
	}
	{
		wxIcon tmp(icon_32_xpm);
		icons.AddIcon(tmp);
	}
	{
		wxIcon tmp(icon_48_xpm);
		icons.AddIcon(tmp);
	}
#endif
	window->SetIcons(icons);
	window->Show(true);
	SetTopWindow(window);
	return true;
}

inline void Brighten(unsigned char& val) {
//	const float temp = 1.0f-((float) val)/255.0f;
//	const float temp2 = sqrt(1.0f-temp*temp);

//	const float temp = ((float) val)/255.0f;
//	const float temp2 = sqrt(temp);
//	val = (unsigned char) (temp2*255.0f);

//	val = (255+val) >> 1;
//	val = (unsigned char) (sqrt((float) val)*16.0f);
#include "gamma_table.h"
	val = gamma[val];
}

template<class Drawer> inline void DrawSimulation(const Sim& sim, Drawer& drawer) {
	if (sim.bm) {
		// draw the selected bot first for great justice (and so the eye cells don't cover other bots)
		drawer.DrawBot(sim.bots.at(sim.bmid), true);
	}
	for (size_t i = sim.bots.size(); i--; ) {
		if (sim.bm && sim.bmid == i) continue;
		const Bot& bot = sim.bots.at(i);
		drawer.DrawBot(bot, false);
		for (shotlist::const_iterator p = bot.shots.begin(); p != bot.shots.end(); ++p) {
			drawer.DrawShot(*p);
		}
	}
	for (shotlist::const_iterator p = sim.orphan_shots.begin(); p != sim.orphan_shots.end(); ++p) {
		drawer.DrawShot(*p);
	}
	for (std::list<Hit>::const_iterator p = sim.hits.begin(); p != sim.hits.end(); ++p) {
		drawer.DrawHit(*p);
	}
}

inline void DrawToolImage(wxDC &dc, wxBitmap bitmap, int x, int y, bool active, bool disabled) {
	wxImage image = bitmap.ConvertToImage();
	if (disabled) {
		image = image.ConvertToGreyscale();
	} else if (active) {
		if (!image.HasAlpha()) image.InitAlpha();
		unsigned char *data = image.GetData();
		for (int i = 24*24; i--; ) {
			Brighten(*data++);
			Brighten(*data++);
			Brighten(*data++);
		}
	}
	bitmap = wxBitmap(image);
	dc.DrawBitmap(bitmap, x, y, true);
}

void wxBotView::OnPaint(wxPaintEvent& event) {
	int w_width, w_height;
	GetClientSize(&w_width, &w_height);

	wxDC* pdc;

	if (IsDoubleBuffered()) {
		pdc = new wxPaintDC(this);
	} else {
		if (bmp_backing == NULL || bmp_backing->GetWidth() != w_width || bmp_backing->GetHeight() != w_height) {
			if (bmp_backing) delete bmp_backing;
			bmp_backing = new wxBitmap(w_width, w_height);
		}
		pdc = new wxBufferedPaintDC(this, *bmp_backing);
	}

//#if wxUSE_GRAPHICS_CONTEXT
//	wxGCDC gdc(*pdc);
//	wxDC& dc = (wxDC&) gdc;
//#else
	wxDC& dc = *pdc;
//#endif

	w_width -= 28;

	dc.SetPen(*wxTRANSPARENT_PEN);
	dc.SetBrush(*wxBLACK_BRUSH);
	dc.DrawRectangle(0, 0, w_width, w_height);

	if (current_sim) {
		const float scale = std::max(((float) current_sim->prefs.width)/((float) w_width),
					((float) current_sim->prefs.height)/((float) w_height))/zoom;
		if (viewx < 0) viewx = 0;
		if (viewy < 0) viewy = 0;
		int maxviewx = current_sim->prefs.width-(w_width*scale);
		int maxviewy = current_sim->prefs.height-(w_height*scale);
		if (maxviewx < 0) maxviewx = 0;
		if (maxviewy < 0) maxviewy = 0;
		if (viewx > maxviewx) viewx = maxviewx;
		if (viewy > maxviewy) viewy = maxviewy;

		// grey out the view
		const bool grey = ((MainFrame*) GetParent())->pause || ((MainFrame*) GetParent())->nodisp;

		//FieldDrawer* draw;
		wxConfigBase* cfg = wxConfigBase::Get();
		switch(cfg->Read(_T("/RenderOptions/Renderer"), RENDERER_WX)) {
			default:
			case RENDERER_WX: {
				WxFieldDrawer draw(dc, viewx, viewy, w_width, w_height, current_sim->prefs.width,
					current_sim->prefs.height, zoom, grey);
				DrawSimulation(*current_sim, draw);
				break;
			}
			case RENDERER_AGG: {
				int back_w, back_h;
				AGGFieldDrawer::BackingSize(w_width, w_height, current_sim->prefs.width,
						current_sim->prefs.height, zoom, back_w, back_h);
				if (img_backing == NULL || img_backing->GetWidth() != back_w || img_backing->GetHeight() != back_h) {
					if (img_backing) delete img_backing;
					img_backing = new wxImage(back_w, back_h, false);
				}
				AGGFieldDrawer draw(dc, *img_backing, viewx, viewy, w_width, w_height, current_sim->prefs.width,
					current_sim->prefs.height, zoom, grey);
				DrawSimulation(*current_sim, draw);
				break;
			}
		}

		//delete draw;

		if (grey) {
			dc.SetPen(*wxTRANSPARENT_PEN);
			dc.SetBrush(wxBrush(wxColor(0, 0, 0), wxHORIZONTAL_HATCH));
			dc.DrawRectangle(0, 0, w_width, w_height);
		}
	}
	// draw toolbar: grey background
	dc.SetPen(*wxTRANSPARENT_PEN);
	dc.SetBrush(*wxMEDIUM_GREY_BRUSH);
	dc.DrawRectangle(w_width, 0, 28, w_height);

	dc.SetPen(*wxBLACK_PEN);
	int y = 2;
#define DrawToolItem(bmp, active, disabled) DrawToolImage(dc, *bmp, w_width+2, y, active, disabled); y += 28
#define DrawToolSeparator() dc.DrawLine(w_width, y, w_width+30, y); y += 3
	DrawToolItem(bmp_new, pressed_item == 0, false);
	DrawToolItem(bmp_load, pressed_item == 1, false);
	DrawToolItem(bmp_save, pressed_item == 2, false);
	DrawToolSeparator();
	DrawToolItem(bmp_play, !((MainFrame*) GetParent())->pause || pressed_item == 3, current_sim == NULL);
	DrawToolItem(bmp_pause, ((MainFrame*) GetParent())->pause || pressed_item == 4, current_sim == NULL);
	DrawToolItem(bmp_cycle, pressed_item == 5, current_sim == NULL || !((MainFrame*) GetParent())->pause);
	DrawToolSeparator();
	DrawToolItem(bmp_nodisplay, ((MainFrame*) GetParent())->nodisp || pressed_item == 6, current_sim == NULL);
	DrawToolSeparator();
	DrawToolItem(bmp_zoomin, pressed_item == 7, current_sim == NULL || zoom > 39.9999f);
	DrawToolItem(bmp_zoomout, pressed_item == 8, current_sim == NULL || zoom < 1.0001f);
	DrawToolItem(bmp_zoomnone, pressed_item == 9, current_sim == NULL || zoom < 1.0001f);
#undef DrawToolItem
#undef DrawToolSeparator

	delete pdc;
}

inline bool TestToolItem(const wxPoint& pos, int w, int tool_y) {
	return (pos.x-w > 0 && pos.x-w < 24 && pos.y > tool_y && pos.y < tool_y+24);
}

void wxBotView::OnDown(wxMouseEvent& event) {
	const wxPoint pos = event.GetPosition();
	int w, h;
	GetClientSize(&w, &h);
	if (pos.x > w || pos.y > h) {
		event.Skip();
		return;
	} else if (pos.x > w-28) {
		w -= 26;
		int y = 2;
		if (TestToolItem(pos, w, y)) pressed_item = 0;
		y += 28;
		if (TestToolItem(pos, w, y)) pressed_item = 1;
		y += 28;
		if (TestToolItem(pos, w, y)) pressed_item = 2;
		y += 31;
		if (TestToolItem(pos, w, y)) pressed_item = 3;
		y += 28;
		if (TestToolItem(pos, w, y)) pressed_item = 4;
		y += 28;
		if (TestToolItem(pos, w, y)) pressed_item = 5;
		y += 31;
		if (TestToolItem(pos, w, y)) pressed_item = 6;
		y += 31;
		if (TestToolItem(pos, w, y)) pressed_item = 7;
		y += 28;
		if (TestToolItem(pos, w, y)) pressed_item = 8;
		y += 28;
		if (TestToolItem(pos, w, y)) pressed_item = 9;
		if (pressed_item != -1) RefreshRect(wxRect(w-2, 0, 28, h), false);
	} else if (current_sim) {
		if (!HasCapture()) CaptureMouse();
		w -= 28;
		const int x = pos.x, y = pos.y;
		const float fltw = static_cast<float>(current_sim->prefs.width), flth = static_cast<float>(current_sim->prefs.height),
			scale = std::max(fltw/static_cast<float>(w), flth/static_cast<float>(h))/zoom;
		const b2Vec2 gpos(x*scale+((float) viewx), y*scale+((float) viewy));
		const b2Vec2 rpos = 0.01f*gpos;
		// test for intersection with a bot
		for (unsigned int i = current_sim->numbots; i--; ) {
			Bot& temp = current_sim->bots.at(i);
			if ((temp.b_body->GetPosition()-rpos).LengthSquared() < temp.radius*temp.radius*0.0001f) {
				current_sim->bm = true;
				current_sim->bmid = i;
				current_sim->down = true;
				current_sim->bmx = (int) ((float) x)*scale+viewx;
				current_sim->bmy = (int) ((float) y)*scale+viewy;
				event.Skip();
				return;
			}
		}
		// not on a bot
		current_sim->bm = false;
		current_sim->down = false;
		mx = (int) x*scale;
		my = (int) y*scale;
	}
	event.Skip();
}

void wxBotView::OnUp(wxMouseEvent& event) {
	if (HasCapture()) ReleaseMouse();
	static const int ids[] = {ID_New, ID_Load, ID_Save, ID_Play, ID_Pause,
		ID_Cycle, ID_NoDisplay, ID_ZoomIn, ID_ZoomOut, ID_ZoomNone};
	const bool disabled[] = {false, false, false, current_sim == NULL,
		current_sim == NULL, current_sim == NULL || !((MainFrame*) GetParent())->pause,
		current_sim == NULL, current_sim == NULL || zoom > 39.9999f, current_sim == NULL || zoom < 1.0001f,
		current_sim == NULL || zoom < 1.0001f};
	if (pressed_item != -1 && pressed_item < 10 && !disabled[pressed_item]) {
		wxCommandEvent evt(wxEVT_COMMAND_TOOL_CLICKED, ids[pressed_item]);
		wxGetApp().window->ProcessEvent(evt);
	}
	pressed_item = mx = my = -1;
	if (current_sim && current_sim->down) {
		current_sim->down = false;
	}
	Refresh(false);
}

void wxBotView::OnDClick(wxMouseEvent& event) {
	const wxPoint pos = event.GetPosition();
	int w, h;
	GetClientSize(&w, &h);
	if (current_sim && current_sim->bm && pos.x < w-28) {
		BotProperties* wnd = new BotProperties(GetParent());
		wnd->SetLabels(current_sim->bots.at(current_sim->bmid));
		wnd->Show();
	} else {
		OnDown(event);
	}
}

void wxBotView::OnRUp(wxMouseEvent& event) {
	if (current_sim) {
		const wxPoint pos = event.GetPosition();
		int w, h;
		GetClientSize(&w, &h);
		w -= 28;
		const int x = pos.x, y = pos.y;
		const float fltw = static_cast<float>(current_sim->prefs.width), flth = static_cast<float>(current_sim->prefs.height),
			scale = std::max(fltw/static_cast<float>(w), flth/static_cast<float>(h))/zoom;
		const b2Vec2 gpos(x*scale+((float) viewx), y*scale+((float) viewy));
		const b2Vec2 rpos = 0.01f*gpos;
		// test for intersection with a bot
		for (unsigned int i = current_sim->numbots; i--; ) {
			Bot& temp = current_sim->bots.at(i);
			if ((temp.b_body->GetPosition()-rpos).LengthSquared() < temp.radius*temp.radius*0.0001f) {
				current_sim->bm = true;
				current_sim->bmid = i;

				return;
			}
		}
	}
}

void wxBotView::OnMove(wxMouseEvent& event) {
	const wxPoint pos = event.GetPosition();
	int w_width, w_height;
	GetClientSize(&w_width, &w_height);
	w_width -= 28;
	if (event.m_leftDown) {
		if (current_sim && mx != -1 && my != -1) {
			// drag view around
			const int x = pos.x, y = pos.y;
			const float fltw = static_cast<float>(current_sim->prefs.width), flth = static_cast<float>(current_sim->prefs.height),
				scale = std::max(fltw/static_cast<float>(w_width), flth/static_cast<float>(w_height))/zoom;
			unsigned int tx = x*scale, ty = y*scale;
			viewx += mx-tx;
			viewy += my-ty;
			mx = tx;
			my = ty;
			if (viewx < 0) {
				mx += viewx;
				viewx = 0;
			}
			if (viewy < 0) {
				my += viewy;
				viewy = 0;
			}
			int maxviewx = current_sim->prefs.width-(w_width*scale);
			int maxviewy = current_sim->prefs.height-(w_height*scale);
			if (maxviewx < 0) maxviewx = 0;
			if (maxviewy < 0) maxviewy = 0;
			if (viewx > maxviewx) {
				mx += viewx-maxviewx;
				viewx = maxviewx;
			}
			if (viewy > maxviewy) {
				my += viewy-maxviewy;
				viewy = maxviewy;
			}
			Refresh(false);
		} else if (pos.x > w_width-28) {
			// change selected item
			w_width += 2;
			int y = 2;
			int old_pressed_item = pressed_item;
			pressed_item = -1;
			if (TestToolItem(pos, w_width, y)) pressed_item = 0;
			y += 28;
			if (TestToolItem(pos, w_width, y)) pressed_item = 1;
			y += 28;
			if (TestToolItem(pos, w_width, y)) pressed_item = 2;
			y += 31;
			if (TestToolItem(pos, w_width, y)) pressed_item = 3;
			y += 28;
			if (TestToolItem(pos, w_width, y)) pressed_item = 4;
			y += 28;
			if (TestToolItem(pos, w_width, y)) pressed_item = 5;
			y += 31;
			if (TestToolItem(pos, w_width, y)) pressed_item = 6;
			y += 31;
			if (TestToolItem(pos, w_width, y)) pressed_item = 7;
			y += 28;
			if (TestToolItem(pos, w_width, y)) pressed_item = 8;
			y += 28;
			if (TestToolItem(pos, w_width, y)) pressed_item = 9;
			if (pressed_item != old_pressed_item) RefreshRect(wxRect(w_width-2, 0, 28, w_height), false);
		} else if (current_sim && current_sim->down) {
			// drag bot around
			const int x = pos.x, y = pos.y;
			const float fltw = static_cast<float>(current_sim->prefs.width), flth = static_cast<float>(current_sim->prefs.height),
				scale = std::max(fltw/static_cast<float>(w_width), flth/static_cast<float>(w_height))/zoom;
			unsigned int tx = x*scale+viewx, ty = y*scale+viewy;
			if (tx > current_sim->prefs.width) tx = current_sim->prefs.width;
			if (ty > current_sim->prefs.height) ty = current_sim->prefs.height;
			current_sim->bmx = tx;
			current_sim->bmy = ty;
		}
	}
}

void wxBotView::OnWheel(wxMouseEvent& event) {
	Zoom(true, pow(1.14471f, ((float) event.GetWheelRotation())/((float) event.GetWheelDelta())));
	Refresh(false);
}

void wxBotView::Zoom(bool in, float amt) {
	if (!current_sim) return; // needs sim for width/height

	int w, h;
	GetClientSize(&w, &h);
	const float fltw = ((float) current_sim->prefs.width), flth = ((float) current_sim->prefs.height);
	const float scale = std::min(fltw/((float) w), flth/((float) h));

	float newzoom = zoom;
	if (in) newzoom *= amt;
	else newzoom /= amt;
	if (!(newzoom != zoom)) return; // returns for NaN and equal
	if (newzoom < 1.0f) newzoom = 1.0f;
	if (newzoom > 40.0f) newzoom = 40.0f;

	// keep zoom center
	const int oldcenter_x = (int) (((float) w)*0.5f*scale/zoom)+viewx, oldcenter_y = (int) (((float) h)*0.5f*scale/zoom)+viewy;
	viewx = oldcenter_x-((int) (scale*((float) w))/(2.0f*newzoom));
	viewy = oldcenter_y-((int) (scale*((float) h))/(2.0f*newzoom));

	zoom = newzoom;
}

void wxBotView::MouseCaptureLost(wxMouseCaptureLostEvent&) {
	while (HasCapture()) ReleaseMouse();
}

MainFrame::MainFrame(const wxString& title, const wxPoint& pos, const wxSize& size)
: wxFrame((wxFrame*) NULL, wxID_ANY, title, pos, size, wxDEFAULT_FRAME_STYLE | wxWS_EX_PROCESS_IDLE),
		pause(false), nodisp(false), saving_or_loading(false) {
	wxMenuBar* menubar = new wxMenuBar;
	wxMenu* menu_file = new wxMenu;
	wxMenu* menu_edit = new wxMenu;
	wxMenu* menu_help = new wxMenu;

	menu_file->Append(ID_New, _T("&New simulation...\tCtrl-N"));
	menu_file->Append(ID_Load, _T("&Load simulation...\tCtrl-O"));
	menu_file->Append(ID_Save, _T("&Save simulation...\tCtrl-S"));
	menu_file->AppendSeparator();
	menu_file->Append(ID_Quit, _T("&Quit\tCtrl-Q"));

	menu_edit->Append(ID_Options, _T("&Preferences...\tCtrl-P"));
	menu_edit->Append(ID_Reset, _T("&Delete all saved config data"));

	menu_help->Append(ID_About, _T("&About"));

	menubar->Append(menu_file, _T("&File"));
	menubar->Append(menu_edit, _T("&Edit"));
	menubar->Append(menu_help, _T("&Help"));

	botview = new wxBotView(this);

	SetMenuBar(menubar);

	int widths[] = {150, 150, -1};

	statusbar = CreateStatusBar();
	statusbar->SetFieldsCount(3, widths);
	statusbar->SetStatusText(_T("No simulation running"), 0);

	// set up idle event handling
	//wxIdleEvent::SetMode(wxIDLE_PROCESS_SPECIFIED);

	/*CreateToolBar();
	wxToolBar* toolbar = GetToolBar();
	toolbar->SetToolBitmapSize(wxSize(24, 24));
	toolbar->AddTool(ID_New, _T("New"), *bmp_new);
	toolbar->AddTool(ID_Load, _T("Load sim"), *bmp_load);
	toolbar->AddTool(ID_Save, _T("Save sim"), *bmp_save);
	toolbar->AddSeparator();
	toolbar->AddTool(ID_Play, _T("Unpause"), *bmp_play, _T(""), wxITEM_RADIO);
	toolbar->AddTool(ID_Pause, _T("Pause"), *bmp_pause, _T(""), wxITEM_RADIO);
	toolbar->AddTool(ID_Cycle, _T("Cycle"), *bmp_cycle);
	toolbar->AddSeparator();
	toolbar->AddTool(ID_NoDisplay, _T("No display"), *bmp_nodisplay, _T(""), wxITEM_CHECK);
	toolbar->AddSeparator();
	toolbar->AddTool(ID_ZoomIn, _T("Zoom in"), *bmp_zoomin);
	toolbar->AddTool(ID_ZoomOut, _T("Zoom out"), *bmp_zoomout);
	toolbar->AddTool(ID_ZoomNone, _T("Reset zoom"), *bmp_zoomnone);
	toolbar->Realize();*/
}

void MainFrame::SaveSim(int compress) {
	if (!current_sim) return;
	saving_or_loading = true;
	wxString filename = wxFileSelector(_T("Save simulation to file.."), _T(""), _T(""), _T(""),
		_T("wxBots sim files (*.wsm)|*.wsm|All files|*.*"), wxFD_SAVE | wxFD_OVERWRITE_PROMPT, this);
	if (!filename.IsEmpty()) {
		wxProgressDialog pdialog(_T("Saving..."), _T("Saving the simulation..."), 200, this,
			wxPD_APP_MODAL | wxPD_CAN_ABORT | wxPD_ELAPSED_TIME | wxPD_REMAINING_TIME);
		std::ofstream ostream(filename.mb_str());
		boost::iostreams::filtering_stream<boost::iostreams::output> fostream;
		fostream.push(boost::iostreams::bzip2_compressor(compress));
		fostream.push(ostream);
		boost::archive::xml_oarchive oarchive(fostream);
		try {
			current_sim->save(oarchive, 0, pdialog);
		} catch (Sim::cancelled) {
		}
	}
	saving_or_loading = false;
}

void MainFrame::OnEvent(wxCommandEvent& event) {
	switch(event.GetId()) {
		case ID_Quit:
			Close(false);
			break;
		case ID_About: {
			wxAboutDialogInfo info;
			info.SetCopyright(_T("Copyright (C) 2008 Geoffry Song"));
			info.SetLicence(_T("wxBots is licensed under the 2-clause BSD license.\n\n"
"Copyright (C) 2008 Geoffry Song\n"
"All rights reserved.\n\n"
"Redistribution and use in source and binary forms, with or without modification,\n"
"are permitted provided that the following conditions are met:\n"
"\n"
"    * Redistributions of source code must retain the above copyright notice,\n"
"this list of conditions and the following disclaimer.\n"
"    * Redistributions in binary form must reproduce the above copyright notice,\n"
"this list of conditions and the following disclaimer in the documentation and/or\n"
"other materials provided with the distribution.\n"
"\n"
"THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS \"AS IS\" AND"
"ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED\n"
"WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE\n"
"DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR\n"
"ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES\n"
"(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;\n"
"LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON\n"
"ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT\n"
"(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS\n"
"SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n\n\n\n"
// Box2D and AGG are usually statically linked in and their licenses require
// stating the copyright
"wxBots uses Box2D, a (2D) physics engine, released under the zlib license.\n\n"
"Copyright (c) 2006-2007 Erin Catto http://www.gphysics.com\n\n"
"This software is provided 'as-is', without any express or implied\n"
"warranty.  In no event will the authors be held liable for any damages\n"
"arising from the use of this software.\n\n"
"Permission is granted to anyone to use this software for any purpose,\n"
"including commercial applications, and to alter it and redistribute it\n"
"freely, subject to the following restrictions:\n\n"
"1. The origin of this software must not be misrepresented; you must not\n"
"   claim that you wrote the original software. If you use this software\n"
"   in a product, an acknowledgment in the product documentation would be\n"
"   appreciated but is not required.\n"
"2. Altered source versions must be plainly marked as such, and must not be\n"
"   misrepresented as being the original software.\n"
"3. This notice may not be removed or altered from any source distribution.\n\n\n\n"
"wxBots uses AGG 2.4, a high-quality rendering engine for C++.\n\n"
"Anti-Grain Geometry - Version 2.4\n"
"Copyright (C) 2002-2004 Maxim Shemanarev (McSeem)\n\n"
"Permission to copy, use, modify, sell and distribute this software\n"
"is granted provided this copyright notice appears in all copies.\n"
"This software is provided \"as is\" without express or implied\n"
"warranty, and with no claim as to its suitability for any purpose.\n\n\n\n"
"wxBots also uses wxWidgets, bzip2, Boost C++ Libraries, licensed under the\n"
"wxWindows Library License, the bzip2 license, and the Boost Software License respectively.\n"));
// whew, that was long
			info.SetVersion(_T("0.0 beta"));
			wxAboutBox(info);
			break;
		}
		case ID_New: {
			NewDialog dlg(this);
			SimSettings prefs;
			dlg.SetFromSettings(prefs);
			if (dlg.ShowModal() == wxID_OK) {
				dlg.SaveToSettings(prefs);
				prefs.Save();
				delete current_sim;
				current_sim = new Sim(prefs);
				botview->zoom = 1.0f;
				botview->viewx = 0;
				botview->viewy = 0;
				botview->Refresh();
			}
			break;
		}
		case ID_Load: {
			// FIXME: load sim
			saving_or_loading = true;
			wxString filename = wxFileSelector(_T("Load simulation from file.."), _T(""), _T(""), _T(""),
				_T("wxBots sim files (*.wsm)|*.wsm|All files|*.*"), wxFD_OPEN | wxFD_FILE_MUST_EXIST, this);
			if (!filename.IsEmpty()) {
				wxProgressDialog pdialog(_T("Loading..."), _T("Loading simulation..."), 200, this,
					wxPD_APP_MODAL | wxPD_CAN_ABORT | wxPD_ELAPSED_TIME | wxPD_REMAINING_TIME);
				std::ifstream istream(filename.mb_str());
				boost::iostreams::filtering_stream<boost::iostreams::input> fistream;
				fistream.push(boost::iostreams::bzip2_decompressor());
				fistream.push(istream);
				boost::archive::xml_iarchive iarchive(fistream);
				//iarchive >> BOOST_SERIALIZATION_NVP(current_sim);
				current_sim = new Sim;
				try {
					current_sim->load(iarchive, 0, pdialog);
					botview->zoom = 1.0f;
					botview->viewx = 0;
					botview->viewy = 0;
					botview->Refresh();
				} catch (Sim::cancelled) {
				}
			}
			saving_or_loading = false;
			break;
		}
		case ID_Save:
			SaveSim();
			break;
		case ID_Options: {
			wxConfigBase* config = wxConfigBase::Get();
			OptionsDialog options(this);
			options.Renderer->SetSelection(config->Read(_T("/RenderOptions/Renderer"), RENDERER_WX));
			if (options.ShowModal() == wxID_OK) {
				config->Write(_T("/RenderOptions/Renderer"), options.Renderer->GetSelection());
			}
			break;
		}
		case ID_Reset:
			wxConfigBase::Get()->DeleteAll();
			wxMessageBox(_T("Settings deleted\n"), _T("Done"), wxOK | wxICON_INFORMATION, this);
			break;
		case ID_Pause:
			pause = true;
			break;
		case ID_Play:
			pause = false;
			break;
		case ID_Cycle:
			if (current_sim && pause) {
				current_sim->cycle();
				statusbar->SetStatusText(wxString::Format(_T("%d cycles elapsed"), current_sim->cycles), 1);
				botview->Refresh(false);
			}
			break;
		case ID_NoDisplay:
			//nodisp = GetToolBar()->GetToolState(ID_NoDisplay);
			nodisp = !nodisp;
			break;
		case ID_ZoomIn:
			botview->Zoom(true, 1.5f);
			break;
		case ID_ZoomOut:
			botview->Zoom(false, 1.5f);
			break;
		case ID_ZoomNone:
			botview->zoom = 1.0f;
			botview->viewx = 0;
			botview->viewy = 0;
			break;
	}
}

void MainFrame::OnClose(wxCloseEvent& event) {
	if (!current_sim) {
		this->Destroy();
	} else switch (wxMessageBox(_T("Save current sim?"), _T("wxBots"), wxYES_NO|wxICON_QUESTION|(event.CanVeto()?wxCANCEL:0), this)) {
		case wxYES:
			// FIXME: save sim
			//current_sim.reset();
			this->Destroy();
			break;
		case wxNO:
			//current_sim.reset();
			this->Destroy();
			break;
		case wxCANCEL:
		default:
			if (!event.CanVeto()) {
				//current_sim.reset();
				this->Destroy();
			} else {
				event.Veto();
				return;
			}
			break;
	}
}

void MainFrame::OnIdle(wxIdleEvent& event) {
	if (!current_sim || pause || saving_or_loading) return;
	cycle = true;
	if (current_sim->cycle()) {
		event.RequestMore(); // om nom nom, idle time
		++cycles_s;
	}
	cycle = false;

	wxString str;
	str << current_sim->cycles;
	str += _T(" cycles elapsed");
	statusbar->SetStatusText(str, 1);
	if (!nodisp) botview->Refresh(false);
	if (stopwatch.Time() > 500) {
		str.Clear();
		str << cycles_s*1000/stopwatch.Time();
		str += _T(" cycles/s");
		statusbar->SetStatusText(str, 0);
		stopwatch.Start();
		cycles_s = 0;
	}
}

